import numpy as np
import numpy as np
import pdb

from abaqus import *
from abaqus import session
from abaqusConstants import *
import __main__
import section
import regionToolset
import displayGroupMdbToolset as dgm
import step
import part
import material
import assembly
import interaction
import load
import mesh
import optimization
import job
import sketch
import visualization
import xyPlot
import displayGroupOdbToolset as dgo
import connectorBehavior
import symbolicConstants
import odbAccess
import shutil

import regionToolset

import csv
from copy import deepcopy


def get_db(db_type):
    """
    Return the current database (either a model or an odb object).

    If a model db is wanted and no model is active, the model in the mdb is selected regardless,
    as long as there is only one model open in the mdb. If no database fits the requirements, None is returned.

    Args:
        db_type: 'odb' or 'model'
    Returns:
        db: database

    NTNU / Knut Andreas Kvaale, 2018
    """
    if db_type is 'model' or db_type is 'mdb':
        if not session_is_odb():
            db = mdb.models[session.viewports['Viewport: 1'].displayedObject.modelName]
        elif len(mdb.models.keys()) is 1:
            db = mdb.models[mdb.models.keys()[0]]
        elif len(mdb.models.keys()) > 1:
            raise AttributeError('No model is not active, and more than one model is available in model database. Impossible to select correct.')
        else:
            db = None
    else:
        if session_is_odb():
            db = session.viewports[session.currentViewportName].displayedObject
        else:
            db = None

    return db

def session_is_odb():
    """
    Check if current session is ODB.

    Returns:
        is_odb: boolean indicating if the session is odb or not
    """    
    is_odb =(('session' in locals() or 'session' in globals()) and
        session.viewports['Viewport: 1'].displayedObject is not None and
        hasattr(session.viewports['Viewport: 1'].displayedObject, 'jobData'))

    return is_odb

def sort_nodes(nodes, sort_nodes_fun=lambda node: node.label):
    if sort_nodes_fun is not None:
        sort_val = [sort_nodes_fun(node) for node in nodes]
        sort_ix = np.argsort(sort_val)
        nodes_new = [None] * len(nodes)
        for node,ix in zip(nodes,sort_ix):
            nodes_new[ix] = node

        return nodes_new
    else:
        return nodes

def get_element_matrices(region, obj=None, sort_nodes_fun=lambda node: node.label):
    
    if region.nodes is None:    #element region
        if obj is None:
            raise ValueError('If region is of element type, `obj` must be given as input. This is either an `instance` object or a `rootAssembly`')

        elements = region.elements
        element_labels = [el.label for el in elements]
        element_matrix = np.zeros([len(element_labels), 3])

        node_labels = []
        for ix,el in enumerate(elements):
            element_matrix[ix,0] = el.label
            nodes = list(el.connectivity)
            if len(nodes)==3:
                __ = nodes.pop(1)    #remove middle node
            element_matrix[ix, 1:] = nodes

            for node in nodes:
                if node not in node_labels:
                    node_labels.append(node)

        nodes = [obj.getNodeFromLabel(node) for node in node_labels]  
        nodes = sort_nodes(nodes, sort_nodes_fun=sort_nodes_fun)
        node_matrix = create_node_matrix(nodes)
        return node_matrix, element_matrix
        
    else: # node region
        if type(region.nodes) is tuple:
            nodes = region.nodes[0]
        else:
            nodes = region.nodes
        nodes = sort_nodes(nodes, sort_nodes_fun=sort_nodes_fun)
        node_matrix = create_node_matrix(nodes)

        return node_matrix


def create_list_of_nodes(region, node_labels):
    return [region.getNodeFromLabel(int(label)) for label in node_labels]

def create_list_of_elements(region, element_labels):
    return [region.getElementFromLabel(int(label)) for label in element_labels]

def create_node_matrix(nodes):
    labels = np.hstack([node.label for node in nodes])[:,np.newaxis]
    coordinates = np.vstack([node.coordinates for node in nodes])
    node_matrix = np.hstack([labels, coordinates])

    return node_matrix

def get_nodal_phi(step_obj, nodes, field_outputs=['U', 'UR'], flatten_components=True, return_base_disp=False):

    if step_obj.domain != MODAL:   #MODAL is a variable in abaqusConstants
        raise TypeError('Type of input step is not modal!')

    n_nodes = len(nodes)
    n_modes = len(step_obj.frames) - 1

    phi = dict()
    basedisp = dict()
    n_dofs = dict()

    for iout, field_output in enumerate(field_outputs):
        foobj0 = step_obj.frames[0].fieldOutputs[field_output]
        
        n_dofs[field_output] = len(foobj0.values[0].data)
        phio = np.zeros([n_dofs[field_output]*n_nodes, n_modes])

        # Get correct data indices to get correct order (as given in node_labels)
        basedisp[field_output] = np.array([foobj0.getSubset(region=node).values[0].data for node in nodes]).flatten()

        for mode in range(0, n_modes):
            foobj = step_obj.frames[mode+1].fieldOutputs[field_output]
            phio[:, mode] = np.array([foobj.getSubset(region=node).values[0].data for node in nodes]).flatten()

        phi[field_output] = phio*1
    
    if flatten_components:
        phi_merged = [None]*n_modes
        
        for mode in range(n_modes):
            phi_merged[mode] = np.hstack([np.array(phi[key][:, mode]).reshape([-1,n_dofs[key]]) for key in field_outputs]).flatten()
        
        basedisp = np.hstack([np.array(basedisp[key]).reshape([-1,n_dofs[key]]) for key in field_outputs]).flatten()
        phi = np.vstack(phi_merged).T

    if return_base_disp:
        return phi, basedisp
    else:
        return phi


def get_element_phi(step_obj, elements, field_outputs=['SF', 'SM'], flatten_components=True, return_integration_points=False):
    if step_obj.domain != MODAL:   #MODAL is a variable in abaqusConstants
        raise TypeError('Type of input step is not modal!')

    n_modes = len(step_obj.frames) - 1
    n_dofs = dict()
    phi = dict()
    integration_points = dict()

    for iout, field_output in enumerate(field_outputs):
        foobj0 = step_obj.frames[0].fieldOutputs[field_output]
        n_dofs[field_output] = len(foobj0.values[0].data)

        # Get correct data indices to get correct order (as given in node_labels)
        all_integration_points = [value.integrationPoint for value in foobj0.values]

        n_int = len(elements) # number of integration points (same element label might appear multiple times if multiple integration points in element)
        phio = np.zeros([n_dofs[field_output]*n_int, n_modes])        

        data_indices = [None]*n_int
        
        for mode in range(0, n_modes):
            foobj = step_obj.frames[mode+1].fieldOutputs[field_output]
            phio[:, mode] = np.array([foobj.getSubset(region=el)[0].data for el in elements]).flatten()

        integration_points[field_output] = [all_integration_points[ix] for ix in data_indices]
        phi[field_output] = phio*1
    
    if flatten_components:
        phi_merged = [None]*n_modes
        
        for mode in range(n_modes):
            phi_merged[mode] = np.hstack([np.array(phi[key][:, mode]).reshape([-1,n_dofs[key]]) for key in field_outputs]).flatten()
            
        integration_points = np.hstack([np.array(integration_points[key]).reshape([-1,n_dofs[key]]) for key in field_outputs]).flatten()
        phi = np.vstack(phi_merged).T


    if return_integration_points:
        return phi, integration_points
    else:
        return phi

    return phi, integration_points


def get_modal_parameters(frequency_step):
    '''
    Output the modal parameters from frequency step of current output database.

    Parameters
    -------------
    frequency_step : str
        name of step containing the modal results (frequency step)

    Returns
    --------------
    f : float
        numpy array with undamped natural frequencies in Hz of all modes computed
    m : float
        numpy array with modal mass for all modes computed
    '''

    odb = get_db('odb')
    history_region_key = odb.steps[frequency_step].historyRegions.keys()[0]

    ftemp = odb.steps[frequency_step].historyRegions[history_region_key].historyOutputs['EIGFREQ'].data
    f = np.array([x[1] for x in ftemp])

    if 'GM' in odb.steps[frequency_step].historyRegions[history_region_key].historyOutputs.keys():
        mtemp = odb.steps[frequency_step].historyRegions[history_region_key].historyOutputs['GM'].data
        m = np.array([x[1] for x in mtemp])
    else:
        m = np.ones(np.shape(f))    #if no GM field is available, mass normalization is assumed used on eigenvalues
    return f, m


import numpy as np
import json
from inspect import isfunction
from scipy.interpolate import interp1d

from wawi.general import blkdiag, interp1d_angular
from wawi.wave import maxincrement, dispersion_relation_scalar, xsim
import wawi.wave
import pyvista as pv
from scipy.linalg import block_diag

'''
HYDRO SUBMODULE
'''     
class Hydro:    
    def __init__(self, pontoons=None, phi_key='hydro', environment=None, phases_lead=False, seastate=None):
            self.pontoons = pontoons
            self.phi_key = phi_key
            self._phases_lead = False
            self._phi = None
            self.seastate = seastate

            if environment is None:
                self.environment = Environment()
            
            self.assign_environment()
 
    @property
    def phases_lead(self):
        return self._phases_lead
    
    @phases_lead.setter
    def phases_lead(self, val):
        self._phases_lead = val
        self.assign_to_pontoon_types(phases_lead=val)
     
    @property
    def phi(self):
        if self._phi is not None:
            return self._phi
        else:
            return np.eye(self.ndofs)
        
    @phi.setter
    def phi(self, phi):
        self._phi = phi

    
    @property
    def pontoon_types(self):
        ptypes = [p.pontoon_type for p in self.pontoons]
        unique_types = []
        for pt in ptypes:
            if pt not in unique_types:
                unique_types.append(pt)
        return unique_types
    
        
    def reset_linearized_drag(self):
        for p in self.pontoons:
            p.Cquad_lin = np.diag(np.zeros(6))
    
    def assign_direct_cquad(self, cquad_dict=None):
        for p in self.pontoons:
            p.Cquad = np.zeros(6)
        
        if cquad_dict is None:
            cquad_dict = {}
            
        for pontoon_label in cquad_dict:
            self.get_pontoon(pontoon_label).Cquad = cquad_dict[pontoon_label]
        
    def assign_environment(self):
        for pontoon in self.pontoons:
            pontoon.environment = self.environment
            
    def assign_to_pontoons(self, **kwargs):
        for p in self.pontoons:
            for k, v in kwargs.items():
                setattr(p, k, v)
    
    def assign_to_pontoon_types(self, **kwargs):
        for pt in self.pontoon_types:
            for k, v in kwargs.items():
                setattr(pt, k, v)
                
    def get_all_cquad(self, local=False):
        
        if local:
            cquad = [p.cquad_local for p in self.pontoons]
        else:
            cquad = [p.cquad for p in self.pontoons]
            
        return block_diag(*cquad)
    
    
        
    def get_all_Cd(self):
        Cd = np.zeros(6*len(self.pontoons))
        for ix, p in enumerate(self.pontoons):
            Cd[ix*6:ix*6+6] = p.Cd
            
        return Cd
    
    def get_all_area(self):
        area = np.zeros(6*len(self.pontoons))
        for ix, p in enumerate(self.pontoons):
            area[ix*6:ix*6+6] = p.area
            
        return area

        
    def get_from_all(self, par):
        vals = []
        for p in self.pontoons:
            vals.append(getattr(p, par))
            
        return vals
                    
    def get_pontoon(self, label):
        matching_pontoons = [p for p in self.pontoons if p.label == label]
        if len(matching_pontoons)>1:
            raise ValueError('Multiple matching pontoons')
        elif len(matching_pontoons)==0:
            return None
        else:
            return matching_pontoons[0]
        
            
    @property
    def nodelabels(self):
        return np.array([pont.node.label for pont in self.pontoons])
    
    @property
    def ndofs(self):
        return len(self.pontoons)*6

'''
ENVIRONMENT CLASS
'''
class Environment:
    def __init__(self, g=9.80655, rho=1.025e3, depth=np.inf, waterlevel=0):
        self.g = g
        self.rho = rho
        self.depth = depth
        self.waterlevel = waterlevel


'''
PONTOON CLASS
'''

class Pontoon:
    def __init__(self, node, pontoon_type, rotation=0, label='pontoon', 
        current_affects_Q=None, seastate=None, environment=None, current_affects_k=None):

        self.node = node
        self.pontoon_type = pontoon_type
        self.rotation = rotation

        self.label = label

        self.seastate = seastate
        self.environment = environment     
        self._current_affects_Q = current_affects_Q
        self._current_affects_k = current_affects_k
        
        self.Cquad = np.zeros(6)   # local, direct quadratic damping
        self._Cd = None


    def __repr__(self):
        return f'Pontoon object: {self.label}'
    def __str__(self):
        return f'Pontoon object: {self.label} at {self.node}'

    def sea_get(self, parameter):
        return self.seastate.get_at_pos(parameter, self.node.x[0], self.node.x[1])


    @property
    def current_affects_Q(self):
        if self._current_affects_Q is not None:
            return self._current_affects_Q
        elif (self.seastate is not None) and ('current_affects_Q' in self.seastate.pontoon_options):
            return self.seastate.pontoon_options['current_affects_Q']
        
    @current_affects_Q.setter
    def current_affects_Q(self, val):
        self._current_affects_Q = val
    
    @property
    def current_affects_k(self):
        if self._current_affects_k is not None:
            return self._current_affects_k
        elif (self.seastate is not None) and ('current_affects_k' in self.seastate.pontoon_options):
            return self.seastate.pontoon_options['current_affects_k']
        
    @current_affects_k.setter
    def current_affects_k(self, val):
        self._current_affects_k = val

    @property
    def Cquad_lin(self):
        if hasattr(self, '_Cquad_lin'):
            return self._Cquad_lin
        else:
            return np.zeros([6, 6])
        
    @Cquad_lin.setter
    def Cquad_lin(self, Cquad_lin):
        self._Cquad_lin = Cquad_lin
        
    @property
    def tmat(self):
        return self.get_tmat_from_rotation()
    
    @property
    def cquad(self):
        T = self.tmat
        return T.T @ self.cquad_local @ T
    
    @property
    def cquad_local(self):
        rho = self.environment.rho
        
        # Directly specified C_quad
        if hasattr(self, 'Cquad'):     #to ensure backwards compatibility
            Cquad = self.Cquad
        else:
            Cquad = np.zeros(6)

        # C_quad from drag definitions
        Cquad_Cd = 0.5 * rho * self.Cd * self.area
        
        return np.diag(Cquad_Cd + Cquad)
        
        
    @property
    def Cd(self):
        if hasattr(self, '_Cd') and self._Cd is not None:
            val = self._Cd
        else:
            val = self.pontoon_type.Cd
            
        if len(val)==3:
            val = np.hstack([val, [0,0,0]])
        else:
            val = np.array(val)
            
        return val
    
    @Cd.setter
    def Cd(self, Cd):
        self._Cd = Cd
    
    @property
    def area(self):
        val = self.pontoon_type.area
            
        if len(val)==3:
            val = np.hstack([val, [0, 0, 0]])
        else:
            val = np.array(val)
            
        return val
    
    # To simplify expressions inside class:
    @property
    def S(self):
        return self.sea_get('S')

    @property
    def D(self):
        return self.sea_get('D')

    @property
    def U(self):
        return self.sea_get('U')

    @property
    def thetaU(self):
        return self.sea_get('thetaU')

    @property
    def depth(self):
        return self.sea_get('depth')

    @property
    def theta0(self):
        return self.sea_get('theta0')

    @property
    def theta_shift(self):
        return self.sea_get('theta_shift')
    
    def doppler_shift_omega(self, omega_k, theta, kappa=None):
        if kappa is None:
            kappa = self.get_kappa(omega_k, theta=theta, U=0.0)

        omega_shifted = theta*0.0
        
        for ix, kappai in enumerate(kappa):
            thetai_abs = theta[ix] + self.theta_shift
            Uproj = self.U * np.cos(self.thetaU - thetai_abs)
            omega_shifted[ix] = omega_k - Uproj * kappai

        return omega_shifted
    
 
    def get_kappa(self, omega_k, theta=0.0, U=None):
        if not self.current_affects_k:
            U = 0.0
        elif U is None:
            U = self.U*1.0

        if U == 0.0:
            kappa = dispersion_relation_scalar(omega_k, self.depth, 
                                               g=self.environment.g)
            
            if np.ndim(theta)!=0:
                kappa = kappa*np.ones(len(theta))
        else:
            kappa = theta*0.0
            for ix, thetai in enumerate(theta):
                thetai_abs = thetai + self.theta_shift
                Uproj = U * np.cos(self.thetaU - thetai_abs)
                kappa[ix] = dispersion_relation_scalar(omega_k, h=self.depth, 
                                                       g=self.environment.g, 
                                                       U=Uproj)

        return kappa
    
    def get_n(self):
        return np.array([np.cos(self.rotation), np.sin(self.rotation), 0])
               
    def get_tmat_from_rotation(self, e2=np.array([0, 0, 1])):
        n = self.get_n()
        e1 = np.cross(e2, n)
        return blkdiag(np.vstack([n, e1, e2]), 2)


    def get_K(self, omega, local=False):
        if local:
            T = np.eye(6)
        else:
            T = self.tmat
            
        if self.pontoon_type.K is not None:
            return T.T @ self.pontoon_type.K(omega) @ T
        else:
            return T*0

        
    def get_C(self, omega,  local=False):
        if local:
            T = np.eye(6)
        else:
            T = self.tmat
            

        return T.T @ (self.pontoon_type.C(omega)) @ T    
    
    
    def get_M(self, omega, local=False):
        if local:
            T = np.eye(6)
        else:
            T = self.tmat
            
        return T.T @ self.pontoon_type.M(omega) @ T   
    
    
    def get_Q0_legacy(self, omega_k, local=False):
        if local:
            T = np.eye(6)
        else:
            T = self.tmat
            
        return T.T @ self.pontoon_type.Q(omega_k)
    
    
    def get_Q0(self, omega, local=False):
        if local:
            T = np.eye(6)
        else:
            T = self.tmat
            
        if np.ndim(omega)==0:    #omega is scalar
            return T.T @ self.pontoon_type.Q(omega)
        else:
            Qeval = self.pontoon_type.Q(omega)
            for k in range(len(omega)):
                Qeval[:,:,k] = T.T @ Qeval[:,:,k]
                
            return Qeval
    
    
    def get_Q(self, omega_k, theta, local=False, theta_interpolation='quadratic'):
        if (not self.current_affects_Q) or (self.U == 0.0):    #omega is scalar
            Qk = interp1d_angular(self.pontoon_type.theta + self.rotation - self.theta_shift, 
                                    self.get_Q0(omega_k, local=local), axis=1, fill_value=0, 
                                    kind=theta_interpolation, bounds_error=False, 
                                    assume_sorted=False)(theta)
            return Qk
        else:
            omega = self.doppler_shift_omega(omega_k, theta)
            Q0 = self.get_Q0(omega, local=local)
            Qk = np.zeros([6, len(theta)]).astype(complex)
            for ix in range(len(theta)):   # TODO: first implementation for loop - will make quicker by avoiding loop later
                Qk[:, ix] = interp1d_angular(self.pontoon_type.theta + self.rotation - self.theta_shift, 
                                        Q0[:, :, ix], axis=1, fill_value=0, 
                                        kind=theta_interpolation, bounds_error=False, 
                                        assume_sorted=False)(theta[ix])
        return Qk     
       
   
    def evaluate_Q(self, omega=None, theta=None, 
                   theta_interpolation='quadratic', local=False):
        
        if omega is None:
            omega = self.pontoon_type.original_omegaQ
        
        if theta is None:
            theta = self.pontoon_type.theta
        
        if np.ndim(theta)==0:
            theta = np.array([theta])

        Q = np.zeros([6, len(theta), len(omega)]).astype('complex')
        
        for k, omega_k in enumerate(omega):
            Q[:, :, k] = self.get_Q(omega_k, theta,
                                    local=local, theta_interpolation=theta_interpolation)
            
        return Q, theta, omega
    
    
    def get_coh_2d(self, omega_k, theta, x0=[0,0], sign=-1):
        kappa = self.get_kappa(omega_k, theta=theta)
        
        # Minus sign in coh confirmed matching with phases lags in OrcaFlex
        
        coh = np.exp(self.pontoon_type.phase_sign*1j*kappa* ((self.node.x[0]-x0[0])*np.cos(theta+self.theta_shift) + 
                                 (self.node.x[1]-x0[1])*np.sin(theta+self.theta_shift)))
        return coh
    
    
    def get_Z(self, omega_k, theta_int, theta_interpolation='quadratic', 
              local=False, x0=[0,0]):   
        
        if theta_int is None:   # => long-crested   
            theta_int = np.array([self.theta0 - self.theta_shift])
                        
        if self.D.__code__.co_argcount==2:    # count number of inputs
            D = self.D(theta_int, omega_k)
        else:
            D = self.D(theta_int)

        # Interpolate hydrodynamic transfer function, get 2d coh and est. Z
        Q_int = self.get_Q(omega_k, theta_int, local=local, 
                           theta_interpolation=theta_interpolation)
        
        coh_2d = self.get_coh_2d(omega_k, theta_int, x0=x0)
        Z = np.sqrt(self.S(omega_k)) * Q_int * np.tile(np.sqrt(D), [6, 1]) * np.tile(coh_2d, [6, 1])
        
        return Z
    
    def get_theta_int(self, omega_k, ds=None, max_rel_error=0.01):
        if ds is None:
            ds = self.max_distance
            
        dtheta = np.min([maxincrement(ds, omega_k, 0, 2*np.pi, max_rel_error), 2*np.pi/20])   #minimum 20 divisions

        return np.arange(-np.pi, np.pi, dtheta)     
    

'''
SEASTATE CLASS
'''

class Seastate:
    def __init__(self, Tp, Hs, gamma, theta0, s=np.inf, depth=None, origin=None, ranges=None, 
                 options={}, pontoon_options={}, name=None, plot_label=None, centered_dirdist=True,
                 U=None, thetaU=None, use_robust_D=True, angle_unit='deg'):

        if angle_unit == 'deg':
            ang_conv = np.pi/180

        if origin is None:
            self.origin = np.array([0,0])
        else:
            self.origin = origin

        if depth is None:
            self.depth = lambda x,y: np.inf
            
        if ranges is None:
            self.ranges = dict(x=[-np.inf, np.inf], y=[-np.inf, np.inf])
        else:
            self.ranges = ranges
            
        self.name = name
        if plot_label is None:
            self.plot_label = name
        else:
            self.plot_label = plot_label
            
        self.pars = ['Tp', 'Hs', 'gamma', 'theta0', 's']
        self.pars_pretty = dict(Tp='$T_p$ [s]', Hs='$H_s$ [m]', gamma=r'$\gamma$', theta0=r'$\theta_0$ $[^\circ]$', s='s')

        if s is None:
            s = np.inf
        if U is None:
            U = 0.0
        if thetaU is None:
            thetaU = 0.0

        self.Tp = Tp
        self.Hs = Hs
        self.gamma = gamma
        self.s = s
        self.theta0 = self.scale(theta0, ang_conv)
        self.centered_dirdist = centered_dirdist
        self.use_robust_D = use_robust_D

        self.U = U
        self.thetaU = self.scale(thetaU, ang_conv)

        self.x0 = self.origin[0]
        self.y0 = self.origin[1]    
        self.options = {'keep_coherence': True}     #default options
        self.options.update(**options)       
        
        self.pontoon_options = {'current_affects_Q': True, 'current_affects_k': True}
        self.pontoon_options.update(**pontoon_options)

        self.fun_pars = [par for par in self.pars if self.isfun(par)]


    def __str__(self):
        string = f'''\
WAWI SeaState 
--------------
Hs={self.Hs:.2f}m
Tp={self.Tp:.2f}s
theta0={self.theta0*180/np.pi:.1f}deg
gamma={self.gamma:.2f}
s={self.s:.1f}
U (current)={self.U:.1f}m/s
thetaU (current)={self.thetaU*180/np.pi:.1f}deg
'''
        
        return string


    @property
    def homogeneous(self):
        if len(self.fun_pars)==0:
            return True         
        else:
            return False
        
    @property
    def theta_shift(self):
        if self.centered_dirdist:
            return self.theta0 
        else:
            return 0.0       
        
    @property
    def short_crested(self):
        return self.s not in [np.inf, None]
    
    @property
    def D(self):
        if not self.short_crested:
            return lambda x, y: lambda theta: 1*(theta==self.evpar('theta0', x, y)-self.evpar('theta_shift',x,y))
        else:
            if self.use_robust_D:
                Dfun = wawi.wave.dirdist_robust
            else:
                Dfun = wawi.wave.dirdist_decimal  
                
            return lambda x,y: Dfun(self.evpar('s', x, y), theta0=self.evpar('theta0', x, y)-self.evpar('theta_shift',x,y))
        
        
    @property
    def S(self):
        return lambda x,y: wawi.wave.jonswap(self.evpar('Hs', x, y), 
                self.evpar('Tp', x, y), self.evpar('gamma', x, y))

    @property
    def theta_int(self):
        if self.short_crested == False:
            th = np.array([self.theta0 - self.theta_shift])
        elif 'truncate_theta' in self.options:
            theta_lims = np.array(self.options['truncate_theta'])
            dtheta = self.options['dtheta']
            th = np.arange(theta_lims[0], theta_lims[1]+dtheta, dtheta)
        else:
            th = None
        
        return th

    def simulate(self, x, y, omega, fs=None, theta=None, phase=None, return_phases=False, 
                 print_progress=True, time_history=True, grid_mode=True):
        
        if theta is None:
            theta = self.theta_int
            
        output = xsim(x, y, self.S, self.D, omega, theta=theta, fs=fs, 
             grid_mode=grid_mode, print_progress=print_progress, time_history=time_history, 
             phase=phase, return_phases=return_phases, theta_shift=self.theta_shift)
        
        if return_phases:
            eta, t, phase = output
            return eta, t, phase
        elif time_history:
            eta, t = output
            return eta, t
        else:
            return output

            

    @staticmethod
    def scale(par, factor):
        if isfunction(par):
            if 'x' in par.__code__.co_varnames and 'y' in par.__code__.co_varnames:
                return lambda x,y: par(x, y)*factor
            elif 'x' in par.__code__.co_varnames:
                return lambda x: par(x)*factor
            elif 'y' in par.__code__.co_varnames:
                return lambda y: par(y)*factor              
        else:
            return par*factor    

    @classmethod
    def from_json(cls, json_file, **kwargs):
        with open(json_file, 'r') as fileobj:
            data = json.load(fileobj)
        
        Tp = eval(data['S']['Tp'])
        Hs = eval(data['S']['Hs'])
        gamma = eval(data['S']['gamma'])
        
        theta0 = eval(data['D']['theta0'])  #assumed in deg
        s = eval(data['D']['s'])

        if 'current' in data:
            U = eval(data['current']['U'])
            thetaU = eval(data['current']['thetaU'])
        else:
            U = None
            thetaU = None
        
        if 'options' in data:
            options = data['options']
        else:
            options = {}
            
        if 'pontoon_options' in data:
            pontoon_options = data['pontoon_options']
        else:
            pontoon_options = {}
        
        if 'origin' in data:
            origin = np.array(data['origin'])
        else:
            origin = np.array([0,0])
            
        if 'ranges' in data:
            ranges = data['ranges']
            if 'x' not in ranges:
                ranges['x'] = [-np.inf, np.inf]
            if 'y' not in ranges:
                ranges['y'] = [-np.inf, np.inf]
        else:
            ranges = None

        if 'name' in data:
            name = data['name']
        else:
            name = 'Unnamed seastate'
        
        if 'plot_label' in data:
            plot_label = data['plot_label']
        else:
            plot_label = None
            
        if 'centered_dirdist' in options:
            centered_dirdist = options.pop('centered_dirdist')
        else:
            centered_dirdist = True

        return cls(Tp, Hs, gamma, theta0, s, origin=origin, ranges=ranges, 
                   name=name, plot_label=plot_label, options=options, 
                   U=U, thetaU=thetaU, pontoon_options=pontoon_options, 
                   centered_dirdist=centered_dirdist,
                   **kwargs)
        
    
    def evpar(self, parameter, x, y):
        
        # Constant extrapolate if outside borders
        if x>np.max(self.ranges['x']): x = np.max(self.ranges['x']) 
        if x<np.min(self.ranges['x']): x = np.min(self.ranges['x'])        
        if y>np.max(self.ranges['y']): y = np.max(self.ranges['y']) 
        if y<np.min(self.ranges['y']): y = np.min(self.ranges['y'])        
        
        par = getattr(self, parameter)
        if isfunction(par):
            if 'x' in par.__code__.co_varnames and 'y' in par.__code__.co_varnames:
                return par(x - self.x0, y - self.y0)
            elif 'x' in par.__code__.co_varnames:
                return par(x - self.x0)
            elif 'y' in par.__code__.co_varnames:
                return par(y - self.y0)                
        else:
            return par
    
    def get_at_pos(self, parameter, x, y):
        par = getattr(self, parameter)
        if isfunction(par):
            if 'x' in par.__code__.co_varnames and 'y' in par.__code__.co_varnames:
                return par(x - self.x0, y - self.y0)
            elif 'x' in par.__code__.co_varnames:
                return par(x - self.x0)
            elif 'y' in par.__code__.co_varnames:
                return par(y - self.y0)                
        else:
            return par
            

    def isfun(self, parameter):
        return isfunction(getattr(self, parameter))

    def with_unit(self, parameter):
        return self.pars_pretty[parameter]


'''
NODE CLASS (MINIMUM CLASS FOR WHEN BEEF MODEL IS NOT APPENDED TO MODEL OBJECT)
'''

class Node:
    def __init__(self, label, x=0.0, y=0.0, z=0.0):
        self.x0 = np.array([x,y,z,0.0,0.0,0.0]).astype(float)
        self.x = np.array([x,y,z,0.0,0.0,0.0]).astype(float)
        self.label = label
        
    def __repr__(self):
        return f'Node {self.label}'
    
    def __str__(self):
        return f'Node {self.label}'
    

'''
PONTOONTYPE CLASS
'''

class PontoonType:
    def __init__(self, K=None, C=None, M=None, Q=None, original_omega=None, original_omegaQ=None,
                 theta=None, label='unnamed', Cd=None, area=None, stl_path=None):
        self.K = K
        self.C = C
        self.M = M
        self.Q = Q
        self.original_omega = original_omega
        self.original_omegaQ = original_omegaQ
        self.theta = theta
        self.label = label
        self.stl_path = stl_path
        
        self.phases_lead = False

        self.generate_mesh()
        self._Cd = Cd
        self._area = area
        
    def __repr__(self):
        return f'PontoonType object: {self.label}'
    def __str__(self):
        return f'PontoonType object: {self.label}'

    def generate_mesh(self):
        if self.stl_path is not None:           
            stl = pv.STLReader(self.stl_path)
            self.mesh = stl.read().translate([0,0,0])
        else:
            self.mesh = None
            
    @property
    def phase_sign(self):
        if hasattr(self, 'phases_lead'):
            return -int(1 - 2*self.phases_lead*1)
        else:
            return -1   #standard is phases lag
    
    @property
    def Cd(self):
        if hasattr(self, '_Cd') and self._Cd is not None:
            return self._Cd
        else:
            return np.array([0,0,0,0,0,0])
        

    @Cd.setter
    def Cd(self, Cd):
        self._Cd = Cd

    @property
    def area(self):
        if hasattr(self, '_area') and self._area is not None:
            return self._area
        else:
            return np.array([0,0,0,0,0,0])
        
    @area.setter
    def area(self, area):
        self._area = area    
    
            
    @classmethod
    def from_numeric(cls, interpolation_kind='linear', fill_value_added='extrapolate', A=None, M0=None, B=None, Kh=None, 
                     omega=None, Q=None, theta=None, omegaQ=None, **kwargs):
        
        if omegaQ is None:
            omegaQ = omega*1
        
        if A is None and M0 is None:
            M = None
        else:
            if A is None:
                A = np.zeros([M0.shape + [len(omega)]])
            elif M0 is None:
                M0 = A[:,:,0]*0
            M = interp1d(omega, (A.T+M0.T).T, axis=2, fill_value=fill_value_added, kind=interpolation_kind)
        
        if B is None:
            C = None
        else:
            C = interp1d(omega, B, axis=2, fill_value=fill_value_added, kind=interpolation_kind)
        
        if Kh is not None:
            K = lambda omega: Kh
        else:
            K = None
        
        if Q is not None:
            Q = interp1d(omegaQ, Q, fill_value=0.0, kind=interpolation_kind, axis=2, bounds_error=False)
    
        return cls(M=M, C=C, K=K, Q=Q, original_omega=omega, original_omegaQ=omegaQ, theta=theta, **kwargs)


    @classmethod
    def from_wadam(cls, path, interpolation_kind='linear', fill_value_added='extrapolate',
                   include=['added mass', 'added damping', 'restoring stiffness', 'inertia'], **kwargs):
        
        from wawi.io import import_wadam_mat, import_wadam_hydro_transfer
        
        A, B, Kh, M0, omega = import_wadam_mat(path)
        omega_Q, theta, Q = import_wadam_hydro_transfer(path)
        
        if 'added mass' not in include:
            A = A*0
        
        if 'added damping' not in include:
            B = B*0
            
        if 'restoring stiffness' not in include:
            Kh = Kh*0
            
        if 'inertia' not in include:
            M0 = M0*0
            
        M = interp1d(omega, (A.T+M0.T).T, axis=2, fill_value=fill_value_added, kind=interpolation_kind)
        C = interp1d(omega, B, axis=2, fill_value=fill_value_added, kind=interpolation_kind)
        K = lambda omega: Kh
        
        Q = interp1d(omega_Q, Q, fill_value=0.0, kind=interpolation_kind, axis=2)
        
        return cls(M=M, C=C, K=K, Q=Q, original_omega=omega, theta=theta, **kwargs)


import numpy as np
import beef
from beef import fe
from copy import deepcopy as copy
import pyvista as pv
from beef.rotation import rodrot, R_from_drot

from pathlib import Path

from ._hydro import Hydro, Pontoon, Environment

from wawi.structural import var_from_modal, expmax_from_modal, peakfactor_from_modal
from wawi.modal import maxreal, normalize_phi, iteig_freq, iteig
from wawi.general import transform_3dmat, eval_3d_fun, fun_const_sum, eval_3d_fun
from wawi.structural import freqsim
from wawi.tools import print_progress as pp
from wawi.wave import stochastic_linearize, harmonic_linearize
from wawi.wind import windaction, windaction_static
# from wawi.io import import_folder

import dill


import matplotlib.pyplot as plt

from scipy.interpolate import interp1d
from scipy.linalg import block_diag
from scipy.signal import savgol_filter

'''
RESULTS SUBMODULE
'''

class Results:
    def __init__(self, psi=None, lambd=None, S=None, omega=None, model=None):
        self.psi = psi
        self.lambd = lambd
        self.S = S
        self.omega = omega
        self.include_eig = None
        self.model = model
        
    @property
    def mdiag(self):
        Kfun, Cfun, Mfun = self.model.get_system_matrices(self.include_eig)
        m = np.zeros(self.lambd.shape[0])
        for ix, wi in enumerate(self.wd):
            m[ix] = np.real(np.diag(np.real(self.psi.T) @ Mfun(wi) @ np.real(self.psi))[ix])
        
        m[m<0] = np.nan
        return m
    
    @property
    def kdiag(self):
        Kfun, Cfun, Mfun = self.model.get_system_matrices(self.include_eig)
        k = np.zeros(self.lambd.shape[0])
        for ix, wi in enumerate(self.wd):
            k[ix] = np.real(np.diag(np.real(self.psi.T) @ Kfun(wi) @ np.real(self.psi))[ix])
            
        k[k<0] = np.nan
        
        return k
    
    @property
    def cdiag(self):
        Kfun, Cfun, Mfun = self.model.get_system_matrices(self.include_eig)
        c = np.zeros(self.lambd.shape[0])
        for ix, wi in enumerate(self.wd):
            c[ix] = np.real(np.diag(np.real(self.psi.T) @ Cfun(wi) @ np.real(self.psi))[ix])       
            
        c[c<0] = np.nan
        
        return c
    
    @property
    def xi(self):
        if self.lambd is not None:
            return -np.real(self.lambd)/np.abs(self.lambd)

    @property
    def wn(self):
        if self.lambd is not None:
            return np.abs(self.lambd)

    @property
    def wd(self):
        if self.lambd is not None:
            return np.abs(np.imag(self.lambd))
   
    @property
    def fn(self):
        if self.lambd is not None:
            return np.abs(self.lambd)/2/np.pi      
        
    @property
    def fd(self):
        if self.lambd is not None:
            return np.abs(np.imag(self.lambd))/2/np.pi                              

    @property
    def Tn(self):
        if self.lambd is not None:
            return 1/(np.abs(self.lambd)/2/np.pi)    
        
    @property
    def Td(self):
        if self.lambd is not None:
            return 1/(np.abs(np.imag(self.lambd))/2/np.pi)   

'''
DRAG ELEMENT CLASS
'''
class DragElement:
    def __init__(self, element, Cd, D=1.0, rho=1025.0, group=None, eltype=None):

        self.group = group
        self._D = D
        self._Cd = Cd
        self.element = element
        self.rho = rho
        self._eltype = eltype
        
    

    @property
    def eltype(self):
        if hasattr(self, '_eltype') and self._eltype is not None:
            return self._eltype
        elif isinstance(self.element, fe.BeamElement3d):
            return 'beam'
        else:
            return 'bar'
        
    @eltype.setter
    def eltype(self, eltype):
        self._eltype = eltype
        

    @property
    def cquad_local(self):
        c = (0.5 * self.Cd * self.rho * self.D)[:3]        #Linearized as 0.25 Cd D L sqrt(8/pi) * stdudot according to A Wang- seems to work well for linearized,check nonlin also
        from beef.general import bar_foundation_stiffness, generic_beam_mat
    
        if self.eltype == 'bar':
            return bar_foundation_stiffness(self.element.L, *(c*0.5))
        elif self.eltype == 'beam':  
            val_dict = {'yy': c[1], 'zz': c[2]}
            return generic_beam_mat(self.element.L, **val_dict)
        elif self.eltype in ['node', 'lumped']:
            C = np.zeros([12,12])
            c = 0.5*c*self.element.L # half on each end
            C[:3,:3] = np.diag(c)
            C[6:6+3,6:6+3] = np.diag(c)
            return C
        else:
            return np.zeros([12,12])
   
    @property
    def cquad(self):                   
        return self.element.tmat.T @ self.cquad_local @ self.element.tmat  
   
    
    def get_cquad_lin(self, var_udot, stochastic=True, 
                      input_is_local=False, ensure_local_output=False):
        # var_udot is covariance matrix of velocity of nodes
        # local_input: whether or not the var_udot is referring to a local csys - if not, it will be transformed to local
        # prior to C computation
        # local_output: whether or not the output should be local - global output is standard
        if not input_is_local:     
            Tin = self.element.tmat
        else:
            Tin = np.eye(12)
            
        if not ensure_local_output:
            Tout = self.element.tmat
        else:
            Tout = np.eye(12)
        
        if stochastic:
            std_udot = np.sqrt(np.abs(Tin @ var_udot @ Tin.T))
            C = stochastic_linearize(self.cquad_local, std_udot)
        else:
            udot = np.diag(np.sqrt(np.abs(Tin @ var_udot @ Tin.T)))    
            C = harmonic_linearize(self.cquad_local, udot)
            
        return Tout.T @ C @ Tout
    
    
    # Get and set methods D and Cd
    @property  
    def D(self):
        if np.ndim(self._D) == 0:
            return np.hstack([self._D*np.ones(3), [0, 0, 0]])
        elif len(self._D) == 3:
            return np.hstack([self._D, [0,0,0]])
        else:
            return self._D
    
    @D.setter
    def D(self, D):
        self._D = D
        
    @property        
    def Cd(self):
        if np.ndim(self._Cd) == 0:
            return np.hstack([self._Cd*np.ones(3), [0,0,0]])
        elif len(self._Cd) == 3:
            return np.hstack([self._Cd, [0,0,0]])
        else:
            return self._Cd
    
    @Cd.setter
    def Cd(self, Cd):
        self._Cd = Cd

    def __repr__(self):
        return f'DragElement <{self.group}> [{self.element.label}]'

    def __str__(self):
        return f'DragElement <{self.group}> [{self.element.label}]'

'''
MODEL CLASS
'''

class Model:
    def __init__(self, hydro=None, aero=None, eldef=None, modal_dry=None, seastate=None, windstate=None,
                 n_dry_modes=None, x0_wave=None, phases_lead=False,
                 use_multibody=True, avoid_eldef=False, drag_elements={}):
        
        self.results = Results(model=self)
        self.hydro = hydro
        self.aero = aero
        self.modal_dry = modal_dry  # ModalDry object
        self.eldef = eldef  # BEEF ElDef object
        self.f_static = None
        
        self.n_dry_modes = n_dry_modes
        
        if modal_dry is not None:
            self.modal_dry.n_modes = n_dry_modes

        if not avoid_eldef and self.eldef is None:
            self.construct_simple_eldef(node_labels=hydro.nodelabels)
        
        if self.aero is not None:
            self.assign_windstate(windstate)

        if self.hydro is not None:
            self.prepare_waveaction()
            self.assign_seastate(seastate)

            if x0_wave is None:
                x,y = self.get_all_pos()
                x0_wave = np.array([np.mean(x), np.mean(y)])
                self.x0 = x0_wave

        if modal_dry is not None:        
            self.assign_dry_modes() 

        self.use_multibody = use_multibody # use multibody if available
        self.phases_lead = phases_lead  # define if phases lead or lag. When lags (lead=False): eta = exp(iwt-kx)
        
        if type(drag_elements) == dict:
            self.assign_drag_elements(drag_elements)
        else:
            self.drag_elements = drag_elements
            
        self.Cquad_lin = 0.0
    
    # Not used due to circular referencing.
    # @staticmethod
    # def from_folder(*args, **kwargs):
    #     '''
    #     Alternative constructor. Passes all inputs to `wawi.io.import_folder`.
    #     '''

    #     model = import_folder(*args, **kwargs)
        
    #     return model
    

    @staticmethod    #alternative constructor
    def from_wwi(path):
        with open(path, 'rb') as f:
            model = dill.load(f)
    
        return model


    def assign_drag_elements(self, drag_elements):
        drag_els = [None]*len(drag_elements)
        for ix, group_key in enumerate(drag_elements):
            els = self.eldef.get_els_with_sec(drag_elements[group_key]['sections'])
            
            # Filter if elements are specified as well
            if hasattr(drag_elements[group_key], 'elements'):
                els = [el for el in els if el in drag_elements[group_key]['elements']]
                
            Cd = drag_elements[group_key]['Cd']
            D = drag_elements[group_key]['D']
            
            if 'rho' in drag_elements[group_key] and drag_elements[group_key]['rho'] is not None:
                rho = drag_elements[group_key]['rho']
            else:
                rho = self.hydro.environment.rho
                
            if 'eltype' in drag_elements[group_key]:
                eltype = drag_elements[group_key]['eltype']
            else:
                eltype = None
            
            drag_els[ix] = [DragElement(el, Cd, D=D, rho=rho, group=group_key, eltype=eltype) for el in els]
        
        self.drag_elements = [a for b in drag_els for a in b]
    
        
    def get_modal_cquad(self, dry=False, psi=None):
        if psi is None:
            if dry:
                psi = np.eye(self.n_modes)
            else:
                psi = self.results.psi
                
        c_quad = np.zeros(self.n_modes)

        if 'full' in self.modal_dry.phi_full and hasattr(self, 'Cquad'):
            phi0 = np.real(self.get_dry_phi(key='full'))
            phi = np.real(phi0 @ psi)
            Cquad_model = self.Cquad
            
            for mode in range(self.n_modes):
                phiabs = np.diag(np.abs(phi[:, mode]))
                c_quad[mode] = c_quad[mode] + phi[:,mode].T @ (Cquad_model * phiabs) @ phi[:,mode]
            
        if 'hydro' in self.modal_dry.phi_full:
            phi0 = self.get_dry_phi(key='hydro')
            phi = np.real(phi0 @ psi)
            Cquad_pontoons = self.hydro.get_all_cquad()
                        
            for mode in range(len(c_quad)):
                phiabs = np.diag(np.abs(phi[:, mode]))
                c_quad[mode] = c_quad[mode] + phi[:,mode].T @ (Cquad_pontoons * phiabs) @ phi[:,mode]
  
        return c_quad
    
    @property
    def tmat_full(self):
        all_tmats = []
        for node in self.eldef.nodes:
            all_tmats.append(self.eldef.get_node_csys(node.label))
        return block_diag(*all_tmats)

    @property
    def Cquad_lin_pontoons(self):    # Only pontoons
        mat = self.initialize_const_matrix()
        for ix, p in enumerate(self.hydro.pontoons):        
            mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.Cquad_lin
        return self.hydro.phi.T @ mat @ self.hydro.phi
    
    
    @property
    def n_modes(self):
        return self.modal_dry.n_modes

    @n_modes.setter
    def n_modes(self, n):
        self.modal_dry.n_modes = n
        self.assign_dry_modes()
        

    @property
    def Cquad(self):    # Only drag elements
        # Requires global representation (phi_full) currently. Consider supporting modal form referring to specific keys.
        # self.local only refers to phi of pontoons
        C = np.zeros([self.eldef.ndofs, self.eldef.ndofs])
        
        if (hasattr(self, 'drag_elements')) and (self.drag_elements is not None) and (len(self.drag_elements) > 0):    
            for e in self.drag_elements:
                 ix = e.element.global_dofs
                 C[np.ix_(ix,ix)] = C[np.ix_(ix,ix)] + e.cquad
             
        return C
        
    def get_Cquad_lin(self, var_udot, local=None, stochastic=True):    # Only drag elements
        
        if local is None:
            local = self.local*1

        if not hasattr(self, 'drag_elements') or len(self.drag_elements) == 0:
            return np.zeros([self.eldef.ndofs, self.eldef.ndofs])
        else:
            C = np.zeros([self.eldef.ndofs, self.eldef.ndofs])
    
        if np.ndim(var_udot)==1:
            var_udot = np.diag(var_udot)
            
        for e in self.drag_elements:
             ix = e.element.global_dofs
             C[np.ix_(ix,ix)] = C[np.ix_(ix,ix)] + e.get_cquad_lin(var_udot[np.ix_(ix,ix)], 
                                                        input_is_local=local, 
                                                        ensure_local_output=local, 
                                                        stochastic=stochastic)    
   
        return C

        
            
    def construct_simple_eldef(self, node_labels=None):
        if node_labels is None:
            node_labels = np.arange(1,len(self.hydro.pontoons)+1,1)
            
        element_matrix = beef.nodes_to_beam_element_matrix(node_labels)
        node_matrix = np.vstack([np.hstack([node_labels[ix], self.hydro.pontoons[ix].node.x0[:3]]) for ix in range(len(node_labels))])
        part = fe.Part(node_matrix, element_matrix)

        for element in part.elements:
            element.assign_e2(np.array([0,1,0]))
            
        part.assign_global_dofs()
        part.update_all_geometry()

        self.eldef = part
        self.connect_eldef()
        
        if 'hydro' in self.modal_dry.phi_full:
            self.modal_dry.phi_full['full'] = self.modal_dry.phi_full['hydro']
    
        
    def initialize_const_matrix(self, astype=float):
        return np.zeros([len(self.hydro.pontoons)*6, len(self.hydro.pontoons)*6]).astype(astype)

    def initialize_freq_matrix(self, n_freq, astype=float):
        return np.zeros([len(self.hydro.pontoons)*6, len(self.hydro.pontoons)*6, n_freq]).astype(astype)

    def __repr__(self):
        return f'WAWI model <{self.__hash__()}>'
    
    def __str__(self):
        return f'WAWI model <{self.__hash__()}>'
    
    def to_wwi(self, path):
        with open(path, 'wb') as f:
            dill.dump(self, f, -1)
        
    
    def get_all_pos(self):
        x = np.array([pont.node.x[0] for pont in self.hydro.pontoons])
        y = np.array([pont.node.x[1] for pont in self.hydro.pontoons])
        return x,y
    
    
    def assign_dry_modes(self):
        if self.hydro is not None:
            self.hydro.phi = self.modal_dry.get_phi(key=self.hydro.phi_key)
        if self.aero is not None:
            self.aero.phi = self.modal_dry.get_phi(key=self.aero.phi_key)


    def connect_eldef(self):
        if self.eldef is not None:
            if self.hydro is not None:
                # Connect pontoons to nodes in beef eldef
                for pontoon in self.hydro.pontoons:
                    pontoon.node = self.eldef.get_node(pontoon.node.label)
            
            # Establish connectivity to aerodynamic sections
            if self.aero is not None:
                self.aero.eldef = dict()
                node_labels_org = self.eldef.get_node_labels()                                   
                
                for group in self.aero.elements:
                    self.aero.eldef[group] = self.eldef.get_element_subset(self.aero.elements[group], renumber=False)  
                    node_labels_sorted = np.array([nl for nl in node_labels_org if nl in self.aero.eldef[group].get_node_labels()])
                    self.aero.eldef[group].arrange_nodes(node_labels_sorted)
                    
                    # These refer to the full dof set, no renumbering yet
                    self.aero.phi_ixs[group] = self.aero.eldef[group].global_dofs*1   

                    # Rearrange dofs based on nodes in subselection, so global dof ixs refer to subset eldef
                    self.aero.eldef[group].assign_node_dofcounts()
                    self.aero.eldef[group].assign_global_dofs() 
                    self.aero.elements[group] = self.aero.eldef[group].elements
                    
                
    def plot_mode(self, mode_ix, use_dry=False, scale=300, title=None, plot_wind_axes=False, 
                  plot_states=['deformed', 'undeformed'], plot_wave_direction=False, **kwargs):
        
        if use_dry: 
            phi_plot = self.get_dry_phi(key='full')
        else:
            phi_plot = np.real(maxreal(self.get_phi(key='full')))
        
        self.eldef.deform_linear(phi_plot[:, mode_ix]*scale, only_deform=True)
        
        if title is None:
            title = f'Mode {mode_ix+1}'
            
        pl = self.plot(plot_states=plot_states, title=title, plot_wind_axes=plot_wind_axes, plot_wave_direction=plot_wave_direction, **kwargs)
        return pl
    
    def export_modeshapes(self, folder, n_modes=None, format='pdf', title=None, zoom=1.0, **kwargs):
        if title is None:
            title = lambda mode: f'Mode {mode+1}:\nTd = {self.results.Td[mode]:.2f}s\nxi = {self.results.xi[mode]*100:.2f}%'
            
        elif type(title) is str:
            title = lambda mode: title + ''

        folder = Path(folder)
        if n_modes is None:
            n_modes = self.n_modes

        for mode in range(n_modes):
            pl = self.plot_mode(mode, show=False, title=title(mode), **kwargs) 
            save_path = folder / f'mode{mode+1:004}.{format}'
            pl.camera.zoom(zoom)
            pl.save_graphic(save_path)


    def copy(self):
        return copy(self)
    
    def plot(self, use_R=False, plot_water=True, 
             waterplane_padding=[1800, 1800], plot_wind_axes=True, wind_ax=[0],
             title=None, show=True, plot_wind_at_centre=True, axis_scaling=100, plot_states=['undeformed'],
             plot_wave_direction=True, wave_origin='center', pontoons_on=['deformed', 'undeformed'],
             thickness_scaling=None, annotate_pontoon_type=False, **kwargs):
        
        if thickness_scaling == 'area':
            lambda sec: np.sqrt(sec.A)
    
        
        tmat_settings = dict(show_edges=False)
        tmat_colors = ['#ff0000', '#00ff00', '#0000ff']
        
        if self.eldef is None:
            pl = pv.Plotter()
        else:
            pl = self.eldef.plot(show=False, plot_states=plot_states,thickness_scaling=thickness_scaling, **kwargs)

        bounds = np.array(pl.bounds)
        origin = (bounds[0::2] + bounds[1::2])/2
        
        if plot_wind_at_centre:
            wind_origin = origin*1
            wind_origin[2] = 10
        else:
            wind_origin = np.array(self.aero.windstate.x_ref)
        
        if self.hydro is not None:
            for pontoon in self.hydro.pontoons:
                if pontoon.pontoon_type.mesh is not None:
                    for state in pontoons_on:
                        x_field = 'x' if state=='deformed' else 'x0'
                        
                        if state in plot_states:
                            # Establish rotation tensor (from global undeformed to global deformed)
                            if use_R and hasattr(pontoon.node, 'R') and state == 'deformed':
                                Tn = pontoon.node.R
                            else:
                                node_rots = getattr(pontoon.node, x_field)[3:]
                                Tn = R_from_drot(node_rots)
                                
                            # Establish rotation tensor (from local to global undeformed)
                            Tzrot = rodrot(pontoon.rotation, style='column')
                            
                            # Stack 4x4 rotation tensor
                            T = np.eye(4)
                            T[:3, :3] = Tn @ Tzrot 
                            
                            mesh_current = copy(pontoon.pontoon_type.mesh)
                            mesh_current = mesh_current.transform(T).translate(getattr(pontoon.node, x_field)[:3])
                            pl.add_mesh(mesh_current)

        # Plot wind arrow
        if plot_wind_axes and self.aero is not None and self.aero.windstate is not None:
            for ax in wind_ax:
                vec = self.aero.windstate.T[ax, :]*axis_scaling
                pl.add_arrows(wind_origin, vec, color=tmat_colors[ax], **tmat_settings)
            
            pl.add_point_labels(np.vstack([wind_origin]), [f'U={self.aero.windstate.U0:.2f} m/s from heading {self.aero.windstate.direction:.1f} deg (CW)'])
            
        # Plot wave arrow        
        if plot_wave_direction and self.hydro is not None and self.hydro.seastate is not None:
            if self.hydro.seastate.homogeneous:
                if wave_origin=='center':
                    wave_origin = origin*1
                    
                vec = rodrot(self.hydro.seastate.theta0)[0,:]*axis_scaling
                pl.add_arrows(np.array(wave_origin-vec*3), vec, color='black', **tmat_settings)
                pl.add_point_labels(np.vstack([wave_origin-vec*3]), [f'dir={self.hydro.seastate.theta0*180/np.pi:.1f} deg, Hs={self.hydro.seastate.Hs:.2f} m, Tp={self.hydro.seastate.Tp:.1f} s'])
            else:
                vec = []
                pos = []
                fun_pars = self.hydro.seastate.fun_pars
                for pontoon in self.hydro.pontoons:
                    pars = dict(theta0=f'theta0 = {pontoon.sea_get("theta0")*180/np.pi:.1f}deg',
                                Hs=f'Hs = {pontoon.sea_get("Hs"):.2f}m',
                                Tp=f'Tp = {pontoon.sea_get("Tp"):.1f}s')
                    
                    vec = rodrot(pontoon.sea_get('theta0'))[0,:]*axis_scaling
                    
                    if wave_origin=='center':
                        wave_origin = origin*1
                        
                    string = ','.join([pars[key] for key in fun_pars])
                    
                    pl.add_arrows(np.array(pontoon.node.x0[:3]+pontoon.tmat[0,:][:3]*200), vec, color='black', **tmat_settings)
                    pl.add_point_labels(np.vstack([pontoon.node.x0[:3]]), [string])
           
        if annotate_pontoon_type:
            for pontoon in self.hydro.pontoons:
                pl.add_point_labels(np.vstack([pontoon.node.x0[:3]]), [pontoon.pontoon_type.label])
       
                    
        # Water plane    
        if plot_water:
            sizes = np.abs(bounds[0::2]-bounds[1::2])
            xsize = sizes[0]+2*waterplane_padding[0]
            ysize = sizes[1]+2*waterplane_padding[1]

            if self.hydro is not None and self.hydro.environment is not None:
                origin[2] = self.hydro.environment.waterlevel
            else:
                origin[2] = 0.0
            water_level = pv.Plane(i_size=xsize, j_size=ysize, center=origin)
            pl.add_mesh(water_level, color='#00aaff', opacity=0.3)
        
        if title is not None:
            pl.add_title(title, color='black', font_size=12)
        
        if show:
            pl.show()
        
        return pl


    @property
    def theta_int(self):
        return self.hydro.seastate.theta_int

    
    @property
    def local(self):
        if self.modal_dry is not None:
            return self.modal_dry.local_phi
        else:
            return False
    
    @property
    def dry_K(self):
        if self.modal_dry is None:
            return 0
        else:
            return self.modal_dry.K

    @property
    def dry_C(self):
        if self.modal_dry is None:
            return 0
        else:
            return self.modal_dry.C

    @property
    def dry_M(self):
        if self.modal_dry is None:
            return 0
        else:
            return self.modal_dry.M

    def get_dry_phi(self, key='hydro'):
        if self.modal_dry is None:
            return np.eye(self.hydro.ndofs)
        else:
            return self.modal_dry.get_phi(key=key)
        
    def get_phi(self, key='hydro', normalize=True, ensure_maxreal=True):
        phi_tot = self.get_dry_phi(key=key) @ self.results.psi
        
        if ensure_maxreal:
            phi_tot = maxreal(phi_tot)
            
        if normalize:
            phi_tot, __ = normalize_phi(phi_tot)
        
        return phi_tot

    def get_result_psd(self, key='hydro', index=None, convert_to=None, modes=None):
        # index: applied after transformation to requested csys (convert_to)

        ix, ix_3d = self.get_mode_ix(modes)            

        if key is not None:    
            sel_phi = self.get_dry_phi(key=key)[:, ix]
        else:
            return self.results.S

        if key in ['hydro', 'full']:   #only supported for hydro and full currently
            if key == 'full':
                tmat = self.tmat_full*1
                if convert_to is not None:
                    print('Local nodal csys is strictly not possible - averaging introduced (use with care).')
            else:
                tmat = self.tmat*1

            if (convert_to == 'global') and (self.local):
                sel_phi = tmat.T @ sel_phi
    
            elif (convert_to == 'local') and (not self.local):
                sel_phi = tmat @ sel_phi

        elif convert_to is not None:
            raise ValueError('convert_to only supported for key="hydro" or "full"; use convert_to=None (output will be given in csys of phi matrix with specified key.')
    
        if index is not None and np.ndim(index)==0:
            index = np.array([index])
            sel_phi = sel_phi[index, :]
        elif index is not None:
            sel_phi = sel_phi[index, :]

        psd = transform_3dmat(self.results.S[ix_3d], sel_phi.T)
        
        if psd.shape[0]==1:
            psd = psd.flatten()
            psd = np.real(psd)
            
        return psd 

    def get_result_std(self, key=None, h=lambda om: 1.0, modes=None):
        ix, ix_3d = self.get_mode_ix(modes)  
 
        if key is None:
            return np.sqrt(np.diag(np.trapz(np.real(self.results.S[ix_3d]*h(self.results.omega)), self.results.omega, axis=2)))
        else:
            return np.sqrt(var_from_modal(self.results.omega, self.results.S[ix_3d]*h(self.results.omega), self.get_dry_phi(key=key)[:,ix]))
    
    def get_result_expmax(self, T, key=None, h=lambda om: 1.0, modes=None):
        ix, ix_3d = self.get_mode_ix(modes)   
        if key is None:
            return expmax_from_modal(self.results.omega, self.results.S[ix_3d]*h(self.results.omega), np.eye(self.results.S.shape[0])[:, ix], T)
        else:
            return expmax_from_modal(self.results.omega, self.results.S[ix_3d]*h(self.results.omega), self.get_dry_phi(key=key)[:,ix], T)  

  
    def get_result_peakfactor(self, T, key='hydro', h=lambda om: 1.0, modes=None):
        ix, ix_3d = self.get_mode_ix(modes)  
        return peakfactor_from_modal(self.results.omega, self.results.S[ix_3d]*h(self.results.omega), self.get_dry_phi(key=key)[:,ix], T)

    def get_gen_S(self, psi=None):
        if psi is None:
            psi = np.real(self.results.psi) # self.results.psi has already been run through maxreal
    
        n_modes = psi.shape[0]

        # CPSD matrices: dry modal response
        Sy = self.results.S*1

        # Estimate Sg
        Sg = Sy*0
        psi_inv = np.linalg.pinv(psi)
        Sg = transform_3dmat(Sy[:n_modes,:n_modes,:], psi_inv.T)

        return Sg
    
    
    def get_mode_ix(self, modes):
        if modes is None:
            modes = np.arange(0, self.results.S.shape[0])
            
        if np.ndim(modes)==0:
            modes = np.array([modes])
        
        return np.array(modes), np.ix_(modes,modes, range(self.results.S.shape[2]))
        
    @property
    def n_pontoons(self):
        return len(self.hydro.pontoons)
    
    @property
    def tmat(self):
        return block_diag(*[pont.tmat for pont in self.hydro.pontoons])


    
    @property
    def pontoon_x(self):
        return np.array([p.node.x[0] for p in self.hydro.pontoons])

    @property
    def pontoon_y(self):
        return np.array([p.node.x[1] for p in self.hydro.pontoons])
    
    @property
    def pontoon_z(self):
        return np.array([p.node.x[2] for p in self.hydro.pontoons])
    

    def plot_2d_tmats(self, scale=50, show_labels=True, ax=None):
        if ax is None:
            fig, ax = plt.subplots()
        
        for pontoon in self.hydro.pontoons:
            x,y,__ = pontoon.node.coordinates
            tmat = pontoon.tmat[:2,:2]*scale
            ex, ey = tmat[0,:], tmat[1,:]
            
            ax.plot([x,x+ex[0]], [y,y+ex[1]], color='blue')
            ax.plot([x,x+ey[0]], [y,y+ey[1]], color='red')

            if show_labels:
                ax.text(x,y,pontoon.label, fontsize=7)
            
        ax.axis('equal')
        return ax

    @classmethod
    def from_nodes_and_types(cls, nodes, pontoon_types, rotation=0, prefix_label='pontoon-', 
                             labels=None, pontoon_props=dict(), **kwargs):     
        if np.ndim(rotation)==0:
            rotation = [rotation]*len(nodes)
            
        pontoons = [None]*len(nodes)
        for ix, node in enumerate(nodes):
            if labels is None:
                label = prefix_label+str(ix+1)
            else:
                label = labels[ix]
            
            pontoons[ix] = Pontoon(node, pontoon_types[ix], rotation=rotation[ix], label=label)            
            
        for pontoon in pontoons:
            for key in pontoon_props:
                setattr(pontoon, key, pontoon_props[key])
        if len(pontoons)>0:
            return cls(hydro=Hydro(pontoons), **kwargs)
        else:
            return cls(**kwargs)
    
            
    def run_eig(self, normalize=False, print_progress=True, freq_kind=True, 
                include=['aero', 'hydro', 'drag_elements', 'drag_pontoons'], 
                smooth=None, aero_sections=None, **kwargs):
        
        include_dict = self.establish_include_dict(include)

        if (('aero' in include_dict['k'] or 'aero' in include_dict['c']) and hasattr(self, 'aero')
            and self.aero is not None and self.aero.Kfun is None):
                self.aero.prepare_aero_matrices(aero_sections=aero_sections)
       
        Kfun, Cfun, Mfun = self.get_system_matrices(include=include_dict)

        if smooth is not None:
            omega = smooth.pop('omega')
            if Kfun is not None:
                K = np.stack([Kfun(omi) for omi in omega], axis=2)
                K = savgol_filter(K, **smooth)
                Kfun = interp1d(omega, K, kind='quadratic')
                
            if Cfun is not None:
                C = np.stack([Cfun(omi) for omi in omega], axis=2)
                C = savgol_filter(C, **smooth)
                Cfun = interp1d(omega, C, kind='quadratic')
                
            if Mfun is not None:
                M = np.stack([Mfun(omi) for omi in omega], axis=2)
                M = savgol_filter(M, **smooth)
                Mfun = interp1d(omega, M, kind='quadratic')
        
        if freq_kind:
            fun = iteig_freq
        else:
            fun = iteig
            
        lambd, psi, __ = fun(Kfun, Cfun, Mfun, print_progress=print_progress, 
                               normalize=normalize, **kwargs)
        
        self.results.psi = maxreal(psi)
        self.results.lambd = lambd
        self.results.include_eig = include


    def run_static(self, aero_sections=None, include_selfexctied=['aero']):
        if not hasattr(self.aero, 'F0_m') or (self.aero.F0_m is None):     # check if already computed static forces
            self.precompute_windaction(static=True, aero_sections=aero_sections)     # compute if not present

        if ('aero' in include_selfexctied) and (self.aero is not None) and (self.aero.Kfun is None):
            K_ae, __ = self.get_aero_matrices(omega_reduced=0, aero_sections=aero_sections)
        else:
            K_ae = 0.0

        Ktot = -K_ae + self.dry_K
        self.results.y_static = np.linalg.inv(Ktot) @ self.aero.F0_m
      

    def run_freqsim(self, omega, omega_aero=None, omega_hydro=None, 
                    print_progress=True, interpolation_kind='linear',
                    max_rel_error=0.01, drag_iterations=0, tol=1e-2, include_action=['hydro', 'aero'],
                    include_selfexcited=['hydro', 'aero', 'drag_elements', 'drag_pontoons'], ensure_at_peaks=True, 
                    theta_interpolation='linear', reset_Cquad_lin=True, aero_sections=None):

        include_dict = self.establish_include_dict(include_selfexcited)

        if (('aero' in include_dict['k'] or 'aero' in include_dict['c']) and hasattr(self, 'aero')
             and (self.aero is not None) and (self.aero.Kfun is None)):
                self.aero.prepare_aero_matrices(omega=omega_aero, aero_sections=aero_sections)

        if reset_Cquad_lin and hasattr(self, 'hydro') and self.hydro is not None:
            self.hydro.reset_linearized_drag()
            self.Cquad_lin = 0.0
        
        if ensure_at_peaks and hasattr(self.results, 'wd') and self.results.wd is not None:
            wd = self.results.wd
            omega_peaks = wd[(wd>=np.min(omega)) & (wd<=np.max(omega))]
            omega = np.unique(np.hstack([omega, omega_peaks]))
        
        if omega_hydro is None:
            omega_hydro = omega*1
        
        if omega_aero is None:
            omega_aero = omega*1
        
        # Excitation
        Sqq_m = 0.0
        
        if (self.hydro is not None) and (self.hydro.seastate is not None) and ('hydro' in include_action):
            if hasattr(self.hydro, 'Sqq_hydro') and (self.hydro.Sqq_hydro is not None):
                Sqq_m = Sqq_m + self.hydro.Sqq_hydro(omega)
            else:
                Sqq_hydro = self.evaluate_waveaction(omega_hydro, print_progress=print_progress, 
                                     max_rel_error=max_rel_error, theta_interpolation=theta_interpolation)
                
                Sqq_m = Sqq_m + interp1d(omega_hydro, Sqq_hydro, kind=interpolation_kind, axis=2, 
                                         fill_value=0.0, bounds_error=False)(omega)
            
        if (self.aero is not None) and (self.aero.windstate is not None) and ('aero' in include_action):
            if hasattr(self.aero, 'Sqq_aero') and (self.aero.Sqq_aero is not None):
                Sqq_m = Sqq_m + self.aero.Sqq_aero(omega)
            else:
                Sqq_aero = self.evaluate_windaction(omega_aero, print_progress=print_progress, aero_sections=aero_sections)
                Sqq_m = Sqq_m + interp1d(omega_aero, Sqq_aero, kind=interpolation_kind, 
                                         axis=2, fill_value=0.0, bounds_error=False)(omega)

        def runsim():
            Hnum = eval_3d_fun(self.get_frf_fun(include=include_dict), omega)  # System
            Srr_m = freqsim(Sqq_m, Hnum) # Power-spectral density method
            
            return Srr_m
        
        Srr_m = runsim()

        # Linearized drag damping           
        if drag_iterations>0:
            Clin_pontoons = np.diag(np.zeros(len(self.hydro.pontoons)*6))   # initialize for convergence check
            
            if ('drag_elements' in include_dict['c']) and hasattr(self, 'Cquad'):
                Cquad_model = self.Cquad * 1
                model_converged = False
                phi_model = self.get_dry_phi(key='full')
                vnodes = np.zeros(phi_model.shape[0])
            else:
                Cquad_model = None
                model_converged = True
                
            if 'drag_pontoons' in include_dict['c']:
                pontoons_converged = False
                vp = np.zeros(Clin_pontoons.shape[0])
                Cquad_pontoons = self.hydro.get_all_cquad()
            else:
                Cquad_pontoons = None
                pontoons_converged = True
            
        for n in range(drag_iterations):
            if print_progress:
                pp(n+1, drag_iterations, postfix=f' DRAG LINEARIZATION - RUNNING ITERATION {n+1}')

            #------ PONTOONS ----------------------------
            if 'drag_pontoons' in include_dict['c']:
                vp_prev = vp * 1
                vp = np.sqrt(var_from_modal(omega, Srr_m*omega**2, self.get_dry_phi(key='hydro')))
                Cquad_lin_pontoons = stochastic_linearize(Cquad_pontoons, vp)
                for ix, p in enumerate(self.hydro.pontoons):
                    p.Cquad_lin = Cquad_lin_pontoons[ix*6:ix*6+6, ix*6:ix*6+6]
                            
                if np.linalg.norm(vp - vp_prev) < tol*(np.max([np.linalg.norm(vp), np.linalg.norm(vp_prev)])):
                    pontoons_converged = True
            
            #------ DRAG ELEMENTS -----------------------
            if ('drag_elements' in include_dict['c']) and (Cquad_model is not None): 
                vnodes_prev = vnodes*1
                var_udot = var_from_modal(omega, Srr_m*omega**2, phi_model, only_diagonal=False)
                self.Cquad_lin = phi_model.T @ self.get_Cquad_lin(var_udot, local=False) @ phi_model
                vnodes = np.sqrt(np.diag(var_udot))
                if np.linalg.norm(vnodes - vnodes_prev) < tol*(np.max([np.linalg.norm(vnodes), np.linalg.norm(vnodes_prev)])):
                    model_converged = True
                
            # Convergence check
            if pontoons_converged and model_converged and print_progress:
                print('\n STOPPING ITERATION. Linearized damping converged by assertion with specified tolerance criterion.')
                break
            
            Srr_m = runsim()
        
        # STORE RESULTS
        self.results.omega = omega*1
        self.results.S = Srr_m*1

    def assign_windstate(self, windstate):
        self.aero.windstate = windstate

    def assign_seastate(self, seastate=None):
        # Reset pontoon settings
        self.hydro.assign_to_pontoons(**dict(current_affects_k=None, current_affects_Q=None))
        self.hydro.Sqq_hydro = None
        
        if seastate is None:
            seastate = self.hydro.seastate
        else:
            self.hydro.seastate = seastate

        for pontoon in self.hydro.pontoons:
            pontoon.seastate = seastate
        
        if seastate is not None:
            self.hydro.assign_to_pontoons(**self.hydro.seastate.pontoon_options)

    def prepare_waveaction(self):
        x, y = self.get_all_pos()
    
        xmesh, ymesh = np.meshgrid(x,x), np.meshgrid(y,y)
        dx = xmesh[0]-xmesh[1]
        dy = ymesh[0]-ymesh[1] 
        ds = np.sqrt(dx**2+dy**2)
        
        min_ix = np.argmin(np.max(ds, axis=0))
        max_distances = ds[min_ix,:] 

        for ix, p in enumerate(self.hydro.pontoons):
            p.max_distance = max_distances[ix]
        
        pont_max_distance = self.hydro.pontoons[np.argmax(max_distances)]
        self.get_theta_int = pont_max_distance.get_theta_int
  
    
       
    def get_waveaction(self, omega_k, max_rel_error=0.01, 
                       theta_interpolation='linear', theta_int=None, transform_by_phi=True):
        
        if theta_int is None and self.theta_int is None:
            theta_int = self.get_theta_int(omega_k, max_rel_error=max_rel_error)
        elif theta_int is None:
            theta_int = self.theta_int
        
        Z = np.zeros([self.hydro.ndofs, len(theta_int)]).astype('complex')
     
        for pontoon_index, pontoon in enumerate(self.hydro.pontoons):                
            Z[pontoon_index*6:pontoon_index*6+6, :] = pontoon.get_Z(omega_k, theta_int, 
                                                                    theta_interpolation=theta_interpolation, 
                                                                    local=self.local, x0=self.x0)
        
        if self.hydro.seastate.short_crested:
            # first and last point in trapezoidal integration has 1/2 as factor, others have 1
            # verified to match for loop over angles and trapz integration.
            dtheta = theta_int[1] - theta_int[0]
            Z[:, 0] = np.sqrt(0.5)*Z[:, 0]
            Z[:, -1] = np.sqrt(0.5)*Z[:, -1]       
            Sqq0 = dtheta * Z @ Z.conj().T
        else:
            Sqq0 = Z @ Z.conj().T
        
        if not self.hydro.seastate.options['keep_coherence']:
            Sqq0 = block_diag(*[Sqq0[i*6:(i+1)*6, i*6:(i+1)*6] for i in range(int(Sqq0.shape[0]/6))])
        
        if transform_by_phi:
            return self.hydro.phi.T @ Sqq0 @ self.hydro.phi
        else:
            return Sqq0

    
    def evaluate_windaction(self, omega=None, aero_sections=None, print_progress=True, static=False, **kwargs):
        if aero_sections is None:
            aero_sections = self.aero.elements.keys()

        Sae_m = 0.0
        T = self.aero.windstate.T
        U = self.aero.windstate.U
        rho = self.aero.windstate.rho

        # Sections needs to be merged - all elements are therefore unwrapped
        els = [a for b in [self.aero.elements[sec] for sec in aero_sections] for a in b]   # all requested sections, flattened
        eldef = self.eldef.get_element_subset(self.eldef.get_elements([el.label for el in els]), renumber=False)    # create new eldef for requested elements   
        phi = self.get_dry_phi(key='full')[eldef.global_dofs, :]    # grab relevant phi components
        eldef.assign_global_dofs()
        nodes = eldef.nodes*1
        els = eldef.elements*1
        
        lc = {sec: self.aero.sections[sec].all_lc for sec in aero_sections}                 # dict with all load coefficients
        B = {sec: self.aero.sections[sec].B for sec in aero_sections}                       # dict with all load coefficients
        D = {sec: self.aero.sections[sec].D for sec in aero_sections}                       # dict with all load coefficients
        S = self.aero.get_generic_kaimal(nodes=nodes)
        section_lookup = {sec: self.aero.elements[sec] for sec in aero_sections}

        if static:
            F0_m = windaction_static(lc, els, T, phi, 
                                B, D, U, print_progress=print_progress, rho=rho,
                                section_lookup=section_lookup, nodes=nodes)  
            return F0_m
        else:
            admittance = {sec: self.aero.sections[sec].admittance for sec in aero_sections}  
            Sae_m_fun = windaction(omega, S, lc, els, T, phi, 
                                B, D, U, print_progress=print_progress, rho=rho,
                                section_lookup=section_lookup, nodes=nodes, admittance=admittance, **kwargs)   
        
            Sae_m = np.stack([Sae_m_fun(om_k) for om_k in omega], axis=2)

            return Sae_m
    
    def evaluate_windaction_static(self, aero_sections=None, print_progress=True, **kwargs):
        return self.evaluate_windaction(aero_sections=None, print_progress=True, static=True, **kwargs)

    def precompute_windaction(self, omega, include=['dynamic'], interpolation_kind='linear', **kwargs):

        if 'dynamic' in include:
            self.aero.Sqq_aero = interp1d(omega, self.evaluate_windaction(omega=omega, static=False, **kwargs), 
                                      kind=interpolation_kind, axis=2, fill_value=0.0, bounds_error=False)
        if 'static' in include:
            self.aero.F0_m = self.evaluate_windaction(static=True, **kwargs)


    def precompute_waveaction(self, omega, interpolation_kind='linear', method='standard', **kwargs):
        if method=='standard':
            Sqq0 = self.evaluate_waveaction(omega, **kwargs)
            
        elif method=='fft':
            if not self.hydro.seastate.homogeneous:
                raise ValueError('Only method standard" is supported for inhomogeneous conditions')
            # Sqq0 = transform_3dmat(waveaction_fft(self.hydro.pontoons, omega, **kwargs), self.hydro.phi)
            raise NotImplementedError('FFT not implemented yet. Will be at some point.')
        
        elif method in ['fourier', 'cos2s', 'cos2s-fourer'] :
            if not self.hydro.homogeneous:
                raise ValueError('Only method standard" is supported for inhomogeneous conditions')
            
            raise NotImplementedError('Fourier (cos 2s) not implemented yet. Will be at some point.')
        
        self.hydro.Sqq_hydro = interp1d(omega, Sqq0, kind=interpolation_kind, axis=2, fill_value=0.0, bounds_error=False)
            


    def evaluate_waveaction(self, omega, max_rel_error=0.01, print_progress=True, theta_int=None,
                            theta_interpolation='quadratic', transform_by_phi=True, **kwargs):
        
        if theta_int is None:
            theta_int = self.theta_int

        if transform_by_phi:
            ndofs = self.hydro.phi.shape[1]
        else:
            ndofs = self.hydro.phi.shape[0]
        
        Sqq = np.zeros([ndofs, ndofs, len(omega)]).astype('complex')

        for k, omega_k in enumerate(omega):
            Sqq[:,:,k] = self.get_waveaction(omega_k, max_rel_error=max_rel_error, theta_int=theta_int, transform_by_phi=transform_by_phi,
                                             theta_interpolation=theta_interpolation, **kwargs)
            
            if print_progress:
                pp(k+1, len(omega), postfix=' ESTABLISHING WAVE EXCITATION     ')
                
        return Sqq
   
    # FRF
    def get_added_frf(self, omega_k, inverse=False):
        if inverse:
            return -omega_k**2*self.get_added_M(omega_k) + 1j*omega_k*self.get_added_C(omega_k) + self.get_added_K(omega_k)
        else:   
            return np.linalg.inv(-omega_k**2*self.get_added_M(omega_k) + 
                                 1j*omega_k*self.get_added_C(omega_k) + self.get_added_K(omega_k))

    def get_dry_frf(self, omega_k, inverse=False):
        if inverse:
            return -omega_k**2*self.dry_M + 1j*omega_k*self.dry_C + self.dry_K
        else:   
            return np.linalg.inv(-omega_k**2*self.dry_M + 1j*omega_k*self.dry_C + self.dry_K)

    def get_aero_K(self, omega_k):
        if self.aero is not None:
            K_aero = self.aero.Kfun(omega_k)
        else:
            K_aero = 0.0
        
        return -K_aero
          
    def get_aero_C(self, omega_k):        
     
        if self.aero is not None:
            C_aero = self.aero.Cfun(omega_k)
        else:
            C_aero = 0.0
            
        return -C_aero
              
            
    def get_aero_M(self, omega_k):
        return 0.0

    def get_hydro_K(self, omega_k):
        if self.hydro is not None:
            mat = self.initialize_const_matrix()
            
            for ix, p in enumerate(self.hydro.pontoons):        
                mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.get_K(omega_k, local=self.local)
            return self.hydro.phi.T @ mat @ self.hydro.phi
        else:
            return 0.0
         
    def get_hydro_C(self, omega_k):        
        if self.hydro is not None:
            mat = self.initialize_const_matrix()
            for ix, p in enumerate(self.hydro.pontoons):        
                mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.get_C(omega_k, local=self.local)
            return self.hydro.phi.T @ mat @ self.hydro.phi
        else:
            return 0.0
              
    
    def get_hydro_M(self, omega_k):
        if self.hydro is not None:
            mat = self.initialize_const_matrix()
            for ix, p in enumerate(self.hydro.pontoons):        
                mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.get_M(omega_k, local=self.local)
    
            return self.hydro.phi.T @ mat @ self.hydro.phi
        else:
            return 0.0

    
    def get_added_K(self, omega_k):
        if self.hydro is not None:
            mat = self.initialize_const_matrix()
            
            for ix, p in enumerate(self.hydro.pontoons):        
                mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.get_K(omega_k, local=self.local)
            K_hydro = self.hydro.phi.T @ mat @ self.hydro.phi
        else:
            K_hydro = 0.0
            
        if self.aero is not None:
            K_aero = self.aero.Kfun(omega_k)
        else:
            K_aero = 0.0
        
        return K_hydro - K_aero
            
    def get_added_C(self, omega_k):        
        if self.hydro is not None:
            mat = self.initialize_const_matrix()
            for ix, p in enumerate(self.hydro.pontoons):        
                mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.get_C(omega_k, local=self.local)
            C_hydro = self.hydro.phi.T @ mat @ self.hydro.phi
        else:
            C_hydro = 0.0
            
        if self.aero is not None:
            C_aero = self.aero.Cfun(omega_k)
        else:
            C_aero = 0.0
            
        return C_hydro - C_aero
              
            
    def get_added_M(self, omega_k):
        if self.hydro is not None:
            mat = self.initialize_const_matrix()
            for ix, p in enumerate(self.hydro.pontoons):        
                mat[ix*6:ix*6+6, ix*6:ix*6+6] = p.get_M(omega_k, local=self.local)
    
            return self.hydro.phi.T @ mat @ self.hydro.phi
        else:
            return 0.0

    @staticmethod
    def establish_include_dict(include):
        if type(include) is not dict:
            include_dict = dict()
            keys = ['k', 'c', 'm']
            for key in keys:
                include_dict[key] = include*1
            return include_dict
        else:
            replacement_keys = dict(stiffness='k', mass='m', damping='c')
            include_dict = {replacement_keys.get(k, k): v for k, v in include.items()}
            return include_dict
            
    def get_system_matrices(self, include=['hydro', 'aero', 'drag_elements', 'drag_pontoons']):
        include_dict = self.establish_include_dict(include)

        # Stiffness
        if ('hydro' in include_dict['k'] and self.hydro is not None) and ('aero' in include_dict['k'] and self.aero is not None):
            Kfun = fun_const_sum(self.get_added_K, self.dry_K)
        elif ('hydro' in include_dict['k'] and self.hydro is not None):
            Kfun = fun_const_sum(self.get_hydro_K, self.dry_K)  
        elif ('aero' in include_dict['k'] and self.aero is not None):
            Kfun = fun_const_sum(self.get_aero_K, self.dry_K)
        else:
            Kfun = lambda omega_k: self.dry_K
            
        # Damping
        dry_C = self.dry_C*1
        if 'drag_elements' in include_dict['c']:
            dry_C += self.Cquad_lin 
        if ('drag_pontoons' in include_dict['c'] and self.hydro is not None):
            dry_C += self.Cquad_lin_pontoons
            
        if ('hydro' in include_dict['c'] and self.hydro is not None) and ('aero' in include_dict['c']  and self.aero is not None):
            Cfun = fun_const_sum(self.get_added_C, dry_C)
        elif ('hydro' in include_dict['c'] and self.hydro is not None):
            Cfun = fun_const_sum(self.get_hydro_C, dry_C)
        elif 'aero' in include_dict['c'] and self.aero is not None:
            Cfun = fun_const_sum(self.get_aero_C, dry_C)
        else:
            Cfun = lambda omega_k: dry_C

        # Mass
        if ('hydro' in include_dict['m'] and self.hydro is not None) and ('aero' in include_dict['m']  and self.aero is not None):
            Mfun = fun_const_sum(self.get_added_M, self.dry_M)
        elif ('hydro' in include_dict['m'] and self.hydro is not None):
            Mfun = fun_const_sum(self.get_hydro_M, self.dry_M)
        elif 'aero' in include_dict['m'] and self.aero is not None: 
            Mfun = fun_const_sum(self.get_aero_M, self.dry_M)
        else:
            Mfun = lambda omega_k: self.dry_M

        return Kfun, Cfun, Mfun
    
    
    def get_frf_fun(self, include=['hydro', 'aero', 'drag_elements'], opt = 0):
        Kfun, Cfun, Mfun = self.get_system_matrices(include)

        def frf(omega_k):
            return np.linalg.inv(-omega_k**2*Mfun(omega_k) + 1j*omega_k*Cfun(omega_k) + Kfun(omega_k))
        
        def imp(omega_k):
            return (-omega_k**2*Mfun(omega_k) + 1j*omega_k*Cfun(omega_k) + Kfun(omega_k))
        if opt == 0:
            return frf
        else:
            return imp
        
    def get_node_ix(self, nodelabel):
        ix = np.where(self.eldef.get_node_labels()==nodelabel)[0]
        if len(ix)>0:
            ix = int(ix[0])
        else:
            ix = None

        return ix
    
    def get_node_ixs(self, nodelabels):
        return [self.get_node_ix(nodelabel) for nodelabel in nodelabels]
    
    def get_aerosection_phi_and_x(self, key):
        ixs = self.aero.phi_ixs[key]
        x = np.vstack([node.coordinates for node in self.aero.eldef[key].nodes])

        return ixs, x




# test wind functions 

# wind state 

# from wawi.model import Windstate

# use the local wawi (Github folder) instead of the installed version
import sys
import os
sys.path.insert(0, os.path.abspath('C:\\Users\\aksef\\Documents\\GitHub\\wawi'))
from wawi.model import Windstate


from math import isclose


def test_windstate():

    V0 = 40.0 # these could be also parametrized

    dir = 45 
    Iu=2.0/V0
    Iw=1.0/V0 
    Au=30 
    Aw=3 
    Cuy=10.0
    Cwy=8.0
    Cuz=10.0
    Cwz=8.0
    Lux=60
    Lwx=60
    x_ref=[0,0,0]
    rho=1.25

    tol = 1e-5

    # Act
    windstate = Windstate(V0, 
                        dir, 
                        Iu=Iu, 
                        Iw=Iw, 
                        Au=Au, Aw=Aw, 
                        Cuy=Cuy, Cwy=Cwy,
                        Cuz=Cuz, Cwz=Cwz,
                        Lux=Lux, Lwx=Lwx,
                        x_ref=x_ref, rho=rho)

    # Assert
    assert isclose(windstate.V0, V0, rel_tol = tol)
    assert isclose(windstate.direction, dir, rel_tol = tol)
    assert isclose(windstate.Iu, Iu, rel_tol = tol)
    assert isclose(windstate.Iw, Iw, rel_tol = tol)
    assert isclose(windstate.Au, Au, rel_tol = tol)
    assert isclose(windstate.Aw, Aw, rel_tol = tol)

    assert isclose(windstate.Cuy, Cuy, rel_tol = tol)
    assert isclose(windstate.Cwy, Cwy, rel_tol = tol)
    assert isclose(windstate.Cuz, Cuz, rel_tol = tol)
    assert isclose(windstate.Cwz, Cwz, rel_tol = tol)

    assert isclose(windstate.Lux, Lux, rel_tol = tol)
    assert isclose(windstate.Lwx, Lwx, rel_tol = tol)
    assert isclose(windstate.rho, rho, rel_tol = tol)

    assert all(abs(a - b) <= tol for a, b in zip(windstate.x_ref, x_ref))









import pytest
import numpy as np
from math import isclose
import dill

# use the local wawi (Github folder) instead of the installed version (remove this when using the installed version)
#import sys
#import os
#sys.path.insert(0, os.path.abspath('C:\\Users\\aksef\\Documents\\GitHub\\wawi'))

# import functions
from wawi.io import import_folder
from wawi.model import Windstate
from wawi.wind import ADs
from wawi.wind import itflutter_cont_naive
from wawi.general import eval_3d_fun

model_folder = './tests/models/model_11c'

   
def AD_dict(AD_funs):
    AD_s = dict(
        A1 = AD_funs['a_fun'][0],
        A2 = AD_funs['a_fun'][1],
        A3 = AD_funs['a_fun'][2],
        A4 = AD_funs['a_fun'][3],
        A5 = AD_funs['a_fun'][4],
        A6 = AD_funs['a_fun'][5],
        
        H1 = AD_funs['h_fun'][0],
        H2 = AD_funs['h_fun'][1],
        H3 = AD_funs['h_fun'][2],
        H4 = AD_funs['h_fun'][3],
        H5 = AD_funs['h_fun'][4],
        H6 = AD_funs['h_fun'][5],
        
        P1 = AD_funs['p_fun'][0],
        P2 = AD_funs['p_fun'][1],
        P3 = AD_funs['p_fun'][2],
        P4 = AD_funs['p_fun'][3],
        P5 = AD_funs['p_fun'][4],
        P6 = AD_funs['p_fun'][5],
        )
    return AD_s


def iabse_11c_windstate(mean_v):
    windstate = Windstate(mean_v, 
                                 90, 
                                 Iu=0.1, 
                                 Iw=0.05, 
                                 Au=6.8, Aw=9.4,  # not used in von Karman
                                 Cuy=0.0, Cwy=0.0,# full coherence assumed
                                 Cuz=0.0, Cwz=0.0,
                                 Lux=200.0, Lwx=20.0,
                                 x_ref=[0,0,0], rho=1.22,
                                 spectrum_type='vonKarman')
    return windstate

davenport = lambda fred: 2*(7*fred-1+np.exp(-7*fred))/(7*fred)**2

@pytest.mark.parametrize(
    'V, expected_f, expected_zeta, tol',
    [(45, np.array([0.0520,	0.1029,	0.2488,]), np.array([0.0142,	0.1660,	0.0429,]), 0.1), 
     (60, np.array([0.0520,	0.1025,	0.2275,]), np.array([0.0169,	0.2879,	0.0366,]), 0.15), 
     (72, np.array([0.0520,	0.0933,	0.2049]), np.array([0.0209,	0.5099,	0.0015,]), 0.25)],
)

def test_in_wind_frequencies_and_damping( V, expected_f, expected_zeta, tol):

    # importing the relevant model
    model = import_folder(model_folder)
    model.modal_dry.xi0 = .3e-2

    # assign ADs (Storebelt ADs)
    with open( model_folder + '/AD_funs_SB_scanlan.pkl', 'rb') as file:
        AD_funs = dill.load(file)
    AD_s = AD_dict(AD_funs)
    model.aero.sections['girder0'].ADs = ADs(**AD_s)

    model.aero.windstate = iabse_11c_windstate(V)
    model.run_eig(w_initial=[0.3, 0.6, 1.5], freq_kind=True, itmax=50)
    lambd = model.results.lambd

    f3 =  np.abs(lambd[2].imag) / (2*np.pi) 
    zeta3 =  -lambd[2].real/ np.abs(lambd[2])

    assert isclose(f3, expected_f[2], rel_tol = tol)
    assert isclose(zeta3, expected_zeta[2], rel_tol = tol)


# flutter speed
def test_flutter():

    expected_flutter_speed = 72.3
    tol = 0.01
    with open( model_folder + '/AD_funs_SB_scanlan.pkl', 'rb') as file:
        AD_funs = dill.load(file)
    ad_dict = AD_dict(AD_funs)

    # importing the relevant model
    model = import_folder(model_folder)
    model.modal_dry.xi0 = .3e-2
    B = model.aero.sections['girder0'].B

    Ms = model.dry_M
    Cs = model.dry_C
    Ks = model.dry_K
    phi = model.get_dry_phi('full')

    # calculate the flutter speed
    res = itflutter_cont_naive(Ms, Cs, Ks, phi, np.array([0,1]), ad_dict, B, V=0.1, rho=1.22, dV=5, 
                overshoot_factor=0.5, itmax={}, tol={}, print_progress=True)

    assert isclose(res['V'][-1], expected_flutter_speed, rel_tol = tol)

# buffeting at 45 m/s (compare with reference values)
def test_mean_speed_45():

    # at 45 m/s & f = 0.278 Hz
    V = 45
    f = 0.278

    # expected values (IABSE report)
    swind_exp = np.array( [ [7.2139, 0], [0, 5.2599] ] )
    # admittance matrix (BqBq matrix in wawi)
    Bq_exp = np.array( [ 
        [	0.0088	,	0.0116	],
        [	0.0076	,	0.2536	],
        [	0.099	,	2.0687	],
    ] )*1e4
    # impedance matrix
    h_exp = np.array( [
        [	-6.74+0.11j	,	-0.01+0.01j	,	-0.29-0.26j	],
        [	0.07+0.04j	,	-6.25+0.80j	,	-19.40-0.37j	],
        [	0.06+0.23j	,	0.50+5.39j	,	-158.52+79.71j	],
    ] )*1e4
    # sign convention
    h_exp[2, :] *= -1
    h_exp[:, 2] *= -1

    # PSD response
    S_exp = np.array( [ [0.26-0j, 1+1j, 1-1j ], 
                        [1-1j, 20-0j, 0-13j ], 
                        [1+1j, 0+13j, 8-0j ], 
                        ])*1e-4
    # sign convention 
    S_exp[2, :] *= -1
    S_exp[:, 2] *= -1

    model = import_folder(model_folder)
    model.aero.windstate = iabse_11c_windstate(V)

    wind_spectrum_func = model.aero.get_generic_kaimal(nodes = model.eldef.nodes)
    wind_spectrum_uu = 2*np.pi*wind_spectrum_func(f*2*np.pi)[0,0]
    wind_spectrum_ww = 2*np.pi*wind_spectrum_func(f*2*np.pi)[2,2]

    model.modal_dry.xi0 = .3e-2
    # assign ADs (Storebelt ADs)
    with open( model_folder + '/AD_funs_SB_scanlan.pkl', 'rb') as file:
        AD_funs = dill.load(file)
    AD_s = AD_dict(AD_funs)
    model.aero.sections['girder0'].ADs = ADs(**AD_s)
    model.aero.prepare_aero_matrices()
    omega = np.array([0.001, f])*np.pi*2
    HH = eval_3d_fun(model.get_frf_fun(opt = 1), omega)

    model.aero.sections['girder0'].admittance = lambda fred: np.full((4, 3), davenport(fred))
    model.run_freqsim(omega,
                    include_selfexcited=['aero'], 
                    include_action=['aero'],
                    print_progress=False)

    global_dof_ix = np.array([1,2,3])
    S = model.get_result_psd(key='full', 
                            index=global_dof_ix)
    
    assert isclose(wind_spectrum_ww, swind_exp[1,1], rel_tol = 0.01)
    assert isclose(wind_spectrum_uu, swind_exp[0,0], rel_tol = 0.01)
    assert np.allclose(np.diag(HH[:,:,1]), np.diag(h_exp), rtol=1e-2)
    assert np.allclose((2*np.pi)*np.diag(S[:,:,-1]), np.diag(S_exp), rtol=5e-2)


@pytest.mark.parametrize(
    'Vbuff, RS_horz_exp, RS_vert_exp, RS_tors_exp',
    [
    (15, 
    np.array([4.00E-02,	2.12E-02,	4.12E+00,	7.00E-05,	6.03E-07,	2.03E-07,	9.11E-08,	3.36E-08,	1.09E-08,]), 
    np.array([6.36E-02,	6.05E-02,	6.64E-02,	9.94E-01,	2.93E-04,	9.17E-05,	1.73E-05,	9.17E-06,	3.41E-06,]),
    np.array([1.52E-03,	1.38E-03,	8.59E-04,	6.95E-05,	2.86E-04,	5.35E-04,	1.16E-02,	5.03E-04,	3.11E-05,]),    
    ), 

    (30, 
    np.array([3.30E-01,	2.76E-01,	8.88E+01,	2.70E-03,	2.84E-05,	1.15E-05,	3.72E-06,	1.85E-06,	6.55E-07,]), 
    np.array([5.89E-01,	5.75E-01,	8.34E-01,	8.70E+00,	1.65E-02,	7.54E-03,	4.73E-04,	2.98E-04,	1.64E-04,]),
    np.array([1.40E-02,	1.33E-02,	1.07E-02,	8.31E-05,	1.60E-02,	4.26E-02,	9.19E-02,	1.40E-02,	1.46E-03,]),    
    ), 

    (45, 
     np.array([1.16E+00,	1.10E+00,	4.02E+02,	1.74E-02,	3.07E-04,	2.02E-04,	2.56E-05,	1.55E-05,	5.92E-06,]), 
     np.array([2.58E+00,	2.55E+00,	3.56E+00,	1.74E+01,	2.18E-01,	2.11E-01,	2.03E-03,	1.45E-03,	1.08E-03,]),
     np.array([6.12E-02,	5.91E-02,	4.50E-02,	2.94E-04,	1.95E-01,	1.11E+00,	1.89E-01,	5.54E-02,	9.36E-03,]),
    ), 

    (60, 
     np.array([3.00E+00,	2.99E+00,	1.01E+03,	6.09E-02,	3.95E-03,	2.53E-03	,8.00E-05,	5.47E-05,	2.47E-05]), 
     np.array([9.43,	9.35,	10.9,	24.7,	2.46,	1.92,	0.00423,	0.00351,	0.00328,]),
     np.array([0.221,	0.215,	0.143,	0.000666,	2,	9.1	, 0.279,	0.115,	0.0272,]),
    ), 
    ]
)

# buffeting responses (spectra)
def test_response_spectra(Vbuff, RS_horz_exp, RS_vert_exp, RS_tors_exp):

    omega_axis = np.array( [0.001, 0.01, 0.052, 0.1, 0.2, 0.234, 0.278, 0.3, 0.35] )*np.pi*2

    # import the model and assign properties
    model = import_folder(model_folder)
    model.aero.windstate = iabse_11c_windstate(Vbuff)
    model.modal_dry.xi0 = .3e-2
    # assign ADs (Storebelt ADs)
    with open( model_folder + '/AD_funs_SB_scanlan.pkl', 'rb') as file:
        AD_funs = dill.load(file)
    AD_s = AD_dict(AD_funs)
    model.aero.sections['girder0'].ADs = ADs(**AD_s)
    model.aero.sections['girder0'].admittance = lambda fred: np.full((4, 3), davenport(fred))

    # run analysis
    model.run_freqsim(omega_axis,
                      include_selfexcited=['aero'], 
                      include_action=['aero'],
                      print_progress=False)
    
    # extract PSD
    global_dof_ix = np.array([1,2,3])
    S = model.get_result_psd(key='full', 
                            index=global_dof_ix)
    
    assert np.allclose((2*np.pi)*S[0,0,:], RS_horz_exp, rtol=5e-2)
    assert np.allclose((2*np.pi)*S[1,1,:], RS_vert_exp, rtol=5e-2)
    assert np.allclose((31/2)**2*(2*np.pi)*S[2,2,:], RS_tors_exp, rtol=5e-2)


# buffeting responses (RMS values)  - for integration we use a finer frequency axis

@pytest.mark.parametrize(
    'Vbuff, RMS_horz_exp, RMS_vert_exp, RMS_tors_exp',
    [(15, 0.0952,	0.173,	0.0211, ), 
     (30, 0.4319,	0.5596,	0.1134,), 
     (45, 0.99,	0.9888,	0.2984, ),  # note that the lateral response is given wrong in the table in the paper. It is taken from the spectral plot.
     (60, 1.7349,	1.5557,	0.7181 ), 
    ])

def test_RMS_response(Vbuff, RMS_horz_exp, RMS_vert_exp, RMS_tors_exp):

    omega_axis = np.linspace(0.001, 6, 1000)

    model = import_folder(model_folder)
    model.aero.windstate = iabse_11c_windstate(Vbuff)
    model.modal_dry.xi0 = .3e-2
    # assign ADs (Storebelt ADs)
    with open( model_folder + '/AD_funs_SB_scanlan.pkl', 'rb') as file:
        AD_funs = dill.load(file)
    AD_s = AD_dict(AD_funs)
    model.aero.sections['girder0'].ADs = ADs(**AD_s)
    model.aero.sections['girder0'].admittance = lambda fred: np.full((4, 3), davenport(fred))
    
    # run analysis
    model.run_freqsim(omega_axis,
                    include_selfexcited=['aero'], 
                    include_action=['aero'],
                    print_progress=False)
    
    # RMS responses
    stds = model.get_result_std(key = 'full')

    assert isclose(stds[1], RMS_horz_exp, rel_tol = 15e-2) # horizontal  # 5-15% std reported in the paper
    assert isclose(stds[2], RMS_vert_exp, rel_tol = 15e-2) # vertical  # 5-15% std reported in the paper
    assert isclose(stds[3]*31/2, RMS_tors_exp, rel_tol = 15e-2) # torsional # 10-20% std reported in the paper



import numpy as np
from scipy.optimize import fsolve

def get_combinations(n_p, n_q=None):
    '''
    Establish all combinations of n indices.

    Parameters
    ----------
    n_p : int
        Number of indices.
    n_q : None, optional
        If integer is given, this is used - otherwise standard value None makes n_q = n_p.

    Returns
    -------
    combos : list of list of int
        List of all unique index combinations (2-by-n_comb).

    Notes
    -----
    Docstring generated using GitHub Copilot.
    '''
    p = np.arange(n_p)
    if n_q is None:
        q = p
        
    combos = []
    for pi in p:
        for qi in q:
            combo = list(np.sort([pi, qi]))
            if (pi != qi) and (combo not in combos):
                combos.append(combo)
                
    return combos

def wave_number_id(Sx, x, k0, Sref=None):
    '''
    Experimental function to establish wave number nonparametrically from cross spectral density.

    Parameters
    ----------
    Sx : ndarray
        Cross spectral density matrix, shape (n_channels, n_channels, n_freq).
    x : ndarray
        Array of spatial positions, shape (n_channels,).
    k0 : float
        Initial guess for wave number (not used in current implementation).
    Sref : None, optional
        Reference spectral density (not used in current implementation).

    Returns
    -------
    k : ndarray
        Estimated wave numbers for each frequency, shape (n_freq,).

    Notes
    -----
    Docstring generated using GitHub Copilot.
    '''
    k = np.zeros(Sx.shape[2])
    n_freq = Sx.shape[2]
    combos = get_combinations(len(x))
             
    b = np.zeros(len(combos)).astype(complex)

    for ix, combo in enumerate(combos):
        dx = x[combo[1]] - x[combo[0]]
        b[ix] = -dx*1j       
        
    for n in range(n_freq):
        lnGamma = np.zeros(len(combos)).astype(complex)
        k_all = [None]*len(combos)
        for ix,combo in enumerate(combos):
            dof1,dof2 = combo
            S = np.sqrt(Sx[dof1,dof1, n]*Sx[dof2,dof2, n])
            
            if S==0:
                lnGamma[ix] = np.nan
            else:
                lnGamma[ix] = np.log(Sx[dof1, dof2, n]/S)
            k_all[ix] = 1j/dx * lnGamma[ix]
            
        k[n] = np.mean(k_all)

        

    return k



import pytest
import numpy as np
from math import isclose

# use the local wawi 
#import sys
#import os
#sys.path.insert(0, r'C:\Users\aksef\OneDrive - Multiconsult\project\wawi\wawi_keep')

# import functions
from wawi.io import import_folder
from wawi.model import Windstate
from wawi.wind import ADs, flatplate_ads
from wawi.wind import itflutter_cont_naive
from wawi.general import eval_3d_fun

model_folder = './tests/models/model_11a'

def iabse_11a_windstate(mean_v):
    windstate = Windstate(mean_v, 
                                    90, 
                                    Iu=0.0, 
                                    Iw=0.05, 
                                    Au=6.8, Aw=9.4, 
                                    Cuy=0.0, Cwy=0.0,
                                    Cuz=0.0, Cwz=0.0,
                                    Lux=0.0, Lwx=20.0,
                                    x_ref=[0,0,0], rho=1.22,
                                    spectrum_type='vonKarman'
                                    )
    return windstate

davenport = lambda fred: 2*(7*fred-1+np.exp(-7*fred))/(7*fred)**2


# in-wind frequencies and damping ratio

@pytest.mark.parametrize(
    'V, expected_f, expected_zeta, tol',
    [(15, np.array([0.0987, 0.279]), np.array([0.0399, 0.0096]), 0.05), 
     (30, np.array([0.0999, 0.2691]), np.array([0.0921, 0.0189]), 0.1), 
     (45, np.array([0.1014, 0.2561]), np.array([0.1689, 0.0309]), 0.1), 
     (60, np.array([0.1027, 0.2340]), np.array([0.3034, 0.0418]), 0.15), 
     (75, np.array([0.0829, 0.1994]), np.array([0.5349, 0.0148]), 0.25)],
)


# aerodynamic stability

# input - wind velocities / output - frequencies, damping ratios ( test only the unstable mode)

def test_in_wind_frequencies_and_damping( V, expected_f, expected_zeta, tol):

    # importing the relevant model
    model = import_folder('./tests/models/model_11a')
    model.modal_dry.xi0 = .3e-2

    # assign flat plate ADs (Theodorsen)
    model.aero.sections['girder0'].ADs = ADs(**flatplate_ads())

    model.aero.windstate = iabse_11a_windstate(V)
    model.run_eig(w_initial=[0.5, 1.25], freq_kind=True, itmax=50)
    lambd = model.results.lambd

    f2 =  np.abs(lambd[1].imag) / (2*np.pi) 
    zeta2 =  -lambd[1].real/ np.abs(lambd[1])

    assert isclose(f2, expected_f[1], rel_tol = tol)
    assert isclose(zeta2, expected_zeta[1], rel_tol = tol)


# flutter speed
def test_flutter():

    expected_flutter_speed = 77.45
    tol = 0.01
    ad_dict = flatplate_ads()

    # importing the relevant model
    model = import_folder(model_folder)
    model.modal_dry.xi0 = .3e-2
    B = model.aero.sections['girder0'].B

    Ms = model.dry_M
    Cs = model.dry_C
    Ks = model.dry_K
    phi = model.get_dry_phi('full')

    # calculate the flutter speed
    res = itflutter_cont_naive(Ms, Cs, Ks, phi, np.array([0,1]), ad_dict, B, V=0.1, rho=1.22, dV=5, 
                overshoot_factor=0.5, itmax={}, tol={}, print_progress=True)

    assert isclose(res['V'][-1], expected_flutter_speed, rel_tol = tol)
    

# buffeting at 45 m/s (compare with analytical)
def test_mean_speed_45():

    # at 45 m/s & f = 0.278 Hz
    V = 45
    f = 0.278

    # expected values (IABSE report)
    sww = 5.26 # wind spectrum
    # impedance matrix
    h_exp = np.array( [[-0.0619+0.0056j, -(-0.1492-0.0811j)], [-(0.01+0.0419j), -1.1559+0.5382j]] )*1e6
    # PSD response
    S_exp = np.transpose(np.array( [ [0.0167, -0.007j], [0.007j, 0.00293] ] ))

    model = import_folder(model_folder)
    model.aero.windstate = iabse_11a_windstate(V)

    wind_spectrum_func = model.aero.get_generic_kaimal(nodes = model.eldef.nodes)
    wind_spectrum_ww = 2*np.pi*wind_spectrum_func(f*2*np.pi)[2,2]

    model.modal_dry.xi0 = .3e-2
    model.aero.sections['girder0'].ADs = ADs(**flatplate_ads())
    model.aero.prepare_aero_matrices()
    omega = np.array([0.001, f])*np.pi*2
    HH = eval_3d_fun(model.get_frf_fun(opt = 1), omega)

    model.aero.sections['girder0'].admittance = lambda fred: np.full((4, 3), davenport(fred))
    model.run_freqsim(omega,
                    include_selfexcited=['aero'], 
                    include_action=['aero'],
                    print_progress=False)

    global_dof_ix = np.array([2,3])
    S = model.get_result_psd(key='full', 
                            index=global_dof_ix)
    
    assert isclose(wind_spectrum_ww, sww, rel_tol = 0.01)
    assert np.allclose(HH[:,:,1], h_exp, rtol=1e-2)
    assert np.allclose((2*np.pi)*S[:,:,-1], S_exp, rtol=5e-2)


# buffeting responses (spectra)

@pytest.mark.parametrize(
    'Vbuff, RS_vert_exp, RS_tors_exp',
    [
    (15, np.array([1.24E-01,	1.19E-01,	1.37E-01,	4.09E+00,	5.66E-04,	1.75E-04,	2.15E-04,	2.18E-05,	7.02E-06,]), 
      np.array([2.53E-03,	2.39E-03,	1.56E-03,	1.85E-04,	4.94E-04,	9.05E-04,	7.07E-02,	1.02E-03,	5.71E-05])   ), 

    (30, np.array([1.19E+00,	1.15E+00,	1.75E+00,	2.06E+01,	3.02E-02,	1.27E-02,	4.61E-03,	9.67E-04,	3.75E-04,]), 
     np.array([2.43E-02,	2.31E-02,	1.98E-02,	9.30E-04,	2.59E-02,	6.16E-02,	4.23E-01,	3.46E-02,	2.94E-03,])  ), 

    (45, np.array([5.60E+00	,	5.40E+00	,	7.45E+00	,	3.43E+01	,	3.41E-01	,	2.50E-01	,	1.67E-02	,	5.85E-03	,	2.78E-03]), 
     np.array([1.15E-01	,	1.08E-01	,	8.46E-02	,	1.55E-03	,	2.85E-01	,	1.11E+00	,	7.06E-01	,	1.52E-01	,	2.07E-02])  ), 

    (60, np.array([2.39E+01	,	2.25E+01	,	2.36E+01	,	4.45E+01	,	3.44E+00	,	6.03E+00	,	3.41E-02	,	1.62E-02	,	9.21E-03,]), 
     np.array([4.89E-01	,	4.51E-01	,	2.67E-01	,	2.00E-03	,	2.77E+00	,	2.38E+01	,	8.28E-01	,	3.06E-01	,	6.37E-02,])  ), 

    (75, np.array([1.47E+02	,	1.24E+02	,	6.17E+01	,	5.32E+01	,	9.48E+02	,	1.79E+00	,	5.34E-02	,	3.04E-02	,	1.98E-02,]), 
     np.array([3.02E+00	,	2.50E+00	,	6.99E-01	,	2.38E-03	,	7.29E+02	,	6.17E+00	,	8.39E-01	,	4.26E-01	,	1.26E-01,])  ), 

     ]
)



# buffeting responses (spectra)
def test_response_spectra(Vbuff, RS_vert_exp, RS_tors_exp):

    omega_axis = np.array( [0.001, 0.01, 0.052, 0.1, 0.2, 0.234, 0.278, 0.3, 0.35] )*np.pi*2

    model = import_folder(model_folder)
    model.aero.windstate = iabse_11a_windstate(Vbuff)
    model.modal_dry.xi0 = .3e-2
    model.aero.sections['girder0'].ADs = ADs(**flatplate_ads())
    model.aero.sections['girder0'].admittance = lambda fred: np.full((4, 3), davenport(fred))
    
    model.run_freqsim(omega_axis,
                    include_selfexcited=['aero'], 
                    include_action=['aero'],
                    print_progress=False, 
                    #merge_aero_sections=True
                    )

    global_dof_ix = np.array([2,3])
    S = model.get_result_psd(key='full', 
                            index=global_dof_ix)

    assert np.allclose((2*np.pi)*S[0,0,:], RS_vert_exp, rtol=5e-2)
    assert np.allclose((31/2)**2*(2*np.pi)*S[1,1,:], RS_tors_exp, rtol=5e-2)



@pytest.mark.parametrize(
    'Vbuff, RMS_vert_exp, RMS_tors_exp',
    [(15, 0.2603, 0.0419  ), 
     (30, 0.778 , 0.2027 ), 
     (45, 1.3404, 0.4792 ), 
     (60, 2.1601, 0.9306 ), 
     (75, 4.4848, 2.8414 )],
)

# buffeting responses (RMS values)  - for integration we use a finer frequency axis
def test_RMS_response(Vbuff, RMS_vert_exp, RMS_tors_exp):

    omega_axis = np.linspace(0.001, 6, 1000)

    model = import_folder(model_folder)
    model.aero.windstate = iabse_11a_windstate(Vbuff)
    model.modal_dry.xi0 = .3e-2
    model.aero.sections['girder0'].ADs = ADs(**flatplate_ads())
    model.aero.sections['girder0'].admittance = lambda fred: np.full((4, 3), davenport(fred))

    model.run_freqsim(omega_axis,
                    include_selfexcited=['aero'], 
                    include_action=['aero'],
                    print_progress=False)
    
    stds = model.get_result_std(key = 'full')

    assert isclose(stds[2], RMS_vert_exp, rel_tol = 10e-2) # vertical  # 5-10% std reported in the paper
    assert isclose(stds[3]*31/2, RMS_tors_exp, rel_tol = 20e-2) # torsional # 10-25% std reported in the paper



import numpy as np
def zero_crossing_period(S, omega):
    """
    Estimate the zero-crossing period from a spectrum.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values.

    Returns
    -------
    float
        Estimated zero-crossing period.
    
    Notes
    -----
    Docstring is generated using Github Copilot.
    """
    return 2*np.pi*np.sqrt(np.trapz(S, x=omega)/np.trapz(omega**2*S, x=omega))

def stoch_mom(S, omega, n=0):
    """
    Compute the n-th spectral moment.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values.
    n : int, optional
        Order of the moment (default is 0).

    Returns
    -------
    float
        n-th spectral moment.
    
    Notes
    -----
    Docstring is generated using Github Copilot.
    """
    return np.trapz(S*omega**n, x=omega)

def m0(S, omega):
    """
    Compute the zeroth spectral moment.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values.

    Returns
    -------
    float
        Zeroth spectral moment.
    
    Notes
    -----
    Docstring is generated using Github Copilot.
    """
    return stoch_mom(S, omega, n=0)

def m2(S, omega):
    """
    Compute the second spectral moment.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values.

    Returns
    -------
    float
        Second spectral moment.
    
    Notes
    -----
    Docstring is generated using Github Copilot.
    """
    return stoch_mom(S, omega, n=2)

def v0_from_spectrum(S, omega):
    """
    Estimate zero-crossing frequency from a spectrum.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values.

    Returns
    -------
    float
        Estimated zero-crossing frequency.
    
    Notes
    -----
    Docstring is generated using Github Copilot.
    """
    return 1/(2*np.pi) * np.sqrt(m2(S, omega)/m0(S, omega))

def v0(m0, m2):
    """
    Calculate zero-crossing frequency from spectral moments.

    Parameters
    ----------
    m0 : float
        Zeroth spectral moment.
    m2 : float
        Second spectral moment.

    Returns
    -------
    float
        Zero-crossing frequency.

    Notes
    -----
    Docstring is generated using Github Copilot.
    """
    return 1/(2*np.pi) * np.sqrt(m2/m0)

def peakfactor(T, v0):
    """
    Calculate the peak factor for a given time period and zero-crossing frequency.

    Parameters
    ----------
    T : float
        The time period over which the peak factor is calculated.
    v0 : float
        The frequency parameter.

    Returns
    -------
    kp : float
        The calculated peak factor.

    Notes
    -----
    The peak factor is computed using the formula:
        c = sqrt(2 * log(v0 * T))
        kp = c + Euler-Mascheroni constant / c
    where `np.euler_gamma` is the Euler-Mascheroni constant.

    Docstring is generated using Github Copilot.

    """

    c = np.sqrt(2*np.log(v0*T))
    kp = c + np.euler_gamma/c
    return kp

def expmax(T, v0, std):
    """
    Calculate the expected maximum value based on temperature, initial value, and standard deviation.

    Parameters
    ----------
    T : float or array-like
        Temperature or an array of temperature values.
    v0 : float
        Initial value parameter.
    std : float
        Standard deviation.

    Returns
    -------
    float or array-like
        The expected maximum value computed as peakfactor(T, v0) multiplied by std.

    Notes
    -----
    This function relies on an external function `peakfactor` which should be defined elsewhere.
    Docstring is generated using Github Copilot.

    Examples
    --------
    >>> expmax(300, 1.0, 0.5)
    1.23  # Example output, depends on peakfactor implementation
    """

    return peakfactor(T,v0)*std

def expmax_from_spectrum(S, omega, T):
    """
    Estimate the expected maximum value from a given spectrum.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values corresponding to the spectrum.
    T : float
        Duration or time interval over which the expected maximum is calculated.

    Returns
    -------
    float
        The expected maximum value estimated from the spectrum.

    Notes
    -----
    This function relies on auxiliary functions `m0`, `m2`, `v0`, and `expmax` to compute
    spectral moments and the expected maximum.

    Docstring is generated using Github Copilot.
    """
    m0_val = m0(S, omega)
    std = np.sqrt(m0_val)

    v0_val = v0(m0_val, m2(S, omega))

    return expmax(T, v0_val, std)

def peakfactor_from_spectrum(S, omega, T):
    """
    Calculate the peak factor from a given spectrum.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    omega : array_like
        Angular frequency values corresponding to the spectrum.
    T : float
        Duration or period over which the peak factor is calculated.

    Returns
    -------
    float
        The calculated peak factor based on the input spectrum and period.

    Notes
    -----
    This function uses the `peakfactor` and `v0` helper functions to compute the result.
    Docstring is generated using Github Copilot.
    """

    return peakfactor(T, v0(S, omega))


import numpy as np

def terrain_roughness(z, z0=0.01, zmin=1.0):
    """
    Calculates the terrain roughness correction for wind speed at a given height.

    Parameters
    ----------
    z : float
        Height above ground level (in meters) at which to calculate the roughness correction.
    z0 : float, optional
        Surface roughness length (in meters). Default is 0.01.
    zmin : float, optional
        Minimum reference height (in meters) to use if `z` is less than or equal to `zmin`. Default is 1.0.

    Returns
    -------
    float
        The terrain roughness correction factor.

    Notes
    -----
    The function uses a logarithmic wind profile and a roughness correction factor `kr` based on the ratio of the provided roughness length `z0` to a reference value `z0_ii = 0.01`. If the height `z` is less than or equal to `zmin`, the calculation uses `zmin` instead of `z`.
    
    References
    ----------
    - EN 1991-1-4: Eurocode 1: Actions on structures – Part 1-4: General actions – Wind actions.
    """

    z0_ii = 0.01
    kr = 0.19 * (z0/z0_ii)**0.07
    
    if z<=zmin:
        return kr*np.log(zmin/z0)
    else:
        return kr*np.log(z/z0)


    
    



from sys import path
from collections import OrderedDict

path.append('C:/Users/knutankv/git-repos/wawi/')   # this is an easy quick fix to enable importing wawi package in Abaqus environment
savefolder = 'C:/Temp/'

# NOTE: Only linear *beam* elements are supported in relevant sets!!

import wawi.ext.abq
import json
import numpy as np

#%% Get database object (ODB) 
db = wawi.ext.abq.get_db('odb')

#%% Definitions
frequency_step = 'Step-11'  # define set name to extract modal properties from
part = db.rootAssembly.instances['PART-1-1']    # define part 
step_obj = db.steps[frequency_step]             

part.ElementSet('ALL', part.elements)   # get all elements (for full element and node matrix)

#%% Grab regions
region_full = part.elementSets['ALL']

#%% Get modal parameters
fn, m = wawi.ext.abq.get_modal_parameters(frequency_step)

#%% Get wind elements and mode shapes
node_matrix, element_matrix = wawi.ext.abq.get_element_matrices(region_full, obj=part)
node_labels = node_matrix[:,0]
nodes = wawi.ext.abq.create_list_of_nodes(part, node_labels)

# Export element definitions as json
el_data = dict(node_matrix=node_matrix.tolist(), 
               element_matrix=element_matrix.tolist())

with open('element.json', 'w') as f:
    json.dump(el_data, f)

phi_full_disp = wawi.ext.abq.get_nodal_phi(step_obj, nodes, flatten_components=True, field_outputs=['U', 'UR'])

## ------------------- EXPORT MODES ----------------
modal_data = dict(omega_n=(fn*2*np.pi).tolist(), m=m.tolist(), phi=dict(full=phi_full_disp.tolist()))

with open('modal.json', 'w') as f:
    json.dump(modal_data, f)